from collections import defaultdict
from functools import partial
from logging import INFO
from typing import Dict, List, Optional, Tuple, Union
import click
from pathlib import Path
from logdecorator import log_on_start
from taf.log import taf_logger
from taf.models.types import Role, RolesIterator
from taf.models.models import TAFKey
from taf.models.types import TargetsRole, MainRoles, UserKeyData
from taf.tuf.repository import MetadataRepository as TUFRepository
from taf.api.utils._conf import find_keystore
from taf.tuf.keys import (
    YkSigner,
    generate_and_write_rsa_keypair,
    generate_rsa_keypair,
    get_sslib_key_from_value,
    load_signer_from_pem,
)

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME, RoleSetupParams
from taf.exceptions import (
    KeystoreError,
    SigningError,
    YubikeyError,
)
from taf.keystore import (
    get_keystore_keys_of_role,
    key_cmd_prompt,
    load_signer_from_private_keystore,
)
from taf import YubikeyMissingLibrary
from securesystemslib.signer._crypto_signer import CryptoSigner

try:
    import taf.yubikey as yk
except ImportError:
    taf_logger.warning(
        "WARNING: yubikey-manager dependency not installed. You will not be able to use YubiKeys."
    )
    yk = YubikeyMissingLibrary()  # type: ignore


def get_key_name(role_name: str, key_num: int, num_of_keys: int) -> str:
    """
    Return a keystore key's name based on the role's name and total number of signing keys,
    as well as the specified counter. If number of signing keys is one, return the role's name.
    If the number of signing keys is greater that one, return role's name + counter (root1, root2...)
    """
    if num_of_keys == 1:
        return role_name
    else:
        return role_name + str(key_num + 1)


def get_metadata_key_info(certs_dir: str, key_id: str) -> TAFKey:
    """
    Read and return information about the specified key read from a certificate
    file whose name matches that key's id.
    """
    cert_path = Path(certs_dir, key_id + ".cert")
    if cert_path.is_file():
        cert_pem = cert_path.read_bytes()
        return TAFKey(key_id, **_extract_x509(cert_pem))

    return TAFKey(key_id)


def _extract_x509(cert_pem: bytes) -> Dict:
    from cryptography import x509
    from cryptography.hazmat.backends import default_backend

    cert = x509.load_pem_x509_certificate(cert_pem, default_backend())

    def _get_attr(oid):
        attrs = cert.subject.get_attributes_for_oid(oid)
        return attrs[0].value if len(attrs) > 0 else ""

    return {
        "name": _get_attr(x509.OID_COMMON_NAME),
        "organization": _get_attr(x509.OID_ORGANIZATION_NAME),
        "country": _get_attr(x509.OID_COUNTRY_NAME),
        "state": _get_attr(x509.OID_STATE_OR_PROVINCE_NAME),
        "locality": _get_attr(x509.OID_LOCALITY_NAME),
        "valid_from": cert.not_valid_before.strftime("%Y-%m-%d"),
        "valid_to": cert.not_valid_after.strftime("%Y-%m-%d"),
    }


def load_sorted_keys_of_new_roles(
    roles: Union[MainRoles, TargetsRole],
    yubikeys_data: Optional[Dict[str, UserKeyData]],
    keystore: Optional[Union[Path, str]],
    yubikeys: Optional[Dict[str, Dict]] = None,
    existing_roles: Optional[List[str]] = None,
    skip_prompt: Optional[bool] = False,
    certs_dir: Optional[Union[Path, str]] = None,
):
    """
    Load signing keys of roles - first those stored on YubiKeys to avoid entering pins
    if there is something wrong with keystore files, then from keystore files.
    Recursively load keys of all delegated roles of target roles.

    Arguments:
        auth_repo: Authentication repository's instance
        roles: MainRoles object (including root, targets, snapshot and timestamp) or a particular delegated role
        yubikeys_data: Contains a mapping of a YubiKey's name to additional details. There names
            are used to link roles to YubiKeys that are used to sign the corresponding metadata.
            If additional details contain the public key, a user will not have to insert that YubiKey
            (provided that it's not necessary given the threshold of signing keys)
        keystore: keystore path
        yubikeys:(optional): A dictionary containing previously loaded YubiKeys used to save already entered pins
        existing_roles (optional): A list of roles whose keys were already loaded
        skip_prompt (optional): A flag defining if the user will be asked if they want to generate new keys or reuse existing
            ones in case keystore files should be used. New keys will be generated by default.
    Side Effects:
        Populates loaded YubiKeys database located in `yubikey.py`

    Returns:
        Signing and verification keys of roles
    """

    def _sort_roles(roles):
        keystore_roles = []
        yubikey_roles = []
        for role in RolesIterator(roles):
            if role.is_yubikey:
                yubikey_roles.append(role)
            else:
                keystore_roles.append(role)
        return keystore_roles, yubikey_roles

    if yubikeys is None:
        yubikeys = defaultdict(dict)
    # load and/or generate all keys first
    if existing_roles is None:
        existing_roles = []
    try:
        keystore_roles, yubikey_roles = _sort_roles(roles)
        signers: Dict = {}
        verification_keys: Dict = {}

        for role in keystore_roles:
            if role.name in existing_roles:
                continue
            keystore_signers, _, _ = setup_roles_keys(
                role,
                keystore=keystore,
                skip_prompt=skip_prompt,
            )
            for signer in keystore_signers:
                signers.setdefault(role.name, []).append(signer)

        for role in yubikey_roles:
            if role.name in existing_roles:
                continue
            _, yubikey_keys, yubikey_signers = setup_roles_keys(
                role,
                certs_dir=certs_dir,
                yubikeys=yubikeys,
                users_yubikeys_details=yubikeys_data,
                skip_prompt=skip_prompt,
            )
            verification_keys[role.name] = yubikey_keys
            signers[role.name] = yubikey_signers
        return signers, verification_keys
    except KeystoreError:
        raise SigningError("Could not load keys of new roles")


def _load_signer_from_keystore(
    taf_repo, keystore_path, key_name, num_of_signatures, scheme, role
) -> CryptoSigner:
    if keystore_path is None:
        return None
    if (keystore_path / key_name).is_file():
        try:
            signer = load_signer_from_private_keystore(
                keystore=keystore_path, key_name=key_name, scheme=scheme
            )
            # load only valid keys
            if taf_repo.is_valid_metadata_key(role, signer.public_key, scheme=scheme):
                return signer
        except KeystoreError:
            pass

    return None


@log_on_start(INFO, "Loading signing keys of '{role:s}'", logger=taf_logger)
def load_signers(
    taf_repo: TUFRepository,
    role: str,
    loaded_yubikeys: Optional[Dict],
    keystore: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
) -> Tuple[List[Dict], List[Dict]]:
    """
    Load role's signing keys. Make sure that at least the threshold of keys was
    loaded, but allow loading more keys (so that a metadata file can be signed
    by all of the role's keys if the user wants that)
    """
    threshold = taf_repo.get_role_threshold(role)
    signing_keys_num = len(taf_repo.get_role_keys(role))
    all_loaded = False
    num_of_signatures = 0
    signers_keystore = []
    yubikeys = []

    # first try to sign using yubikey
    # if that is not possible, try to load key from a keystore file
    # if the keystore file is not found, ask the user if they want to sign
    # using yubikey and to insert it if that is the case

    if keystore is None:
        keystore_path = find_keystore(taf_repo.path)
        if keystore_path is None:
            taf_logger.warning("No keystore provided and no default keystore found")
        else:
            keystore = str(keystore_path)

    keystore_path = Path(keystore).expanduser().resolve() if keystore else None

    def _load_and_append_yubikeys(
        key_name, role, retry_on_failure, hide_already_loaded_message
    ):
        public_key, serial_num = yk.yubikey_prompt(
            key_name,
            role,
            taf_repo,
            loaded_yubikeys=loaded_yubikeys,
            retry_on_failure=retry_on_failure,
            hide_already_loaded_message=hide_already_loaded_message,
        )
        if public_key is not None and public_key not in yubikeys:
            signer = YkSigner(
                public_key, partial(yk.yk_secrets_handler, serial_num=serial_num)
            )
            yubikeys.append(signer)
            taf_logger.info(f"Successfully loaded {key_name} from inserted YubiKey")
            return True
        return False

    keystore_files = []
    if keystore is not None:
        keystore_files = get_keystore_keys_of_role(keystore, role)
    prompt_for_yubikey = True
    use_yubikey_for_signing_confirmed = False
    while not all_loaded and num_of_signatures < signing_keys_num:

        # when loading from keystore files
        # there is no need to ask the user if they want to load more key, try to load from keystore
        if num_of_signatures < len(keystore_files):
            key_name = keystore_files[num_of_signatures]
            signer = _load_signer_from_keystore(
                taf_repo, keystore_path, key_name, num_of_signatures, scheme, role
            )
            if signer is not None:
                signers_keystore.append(signer)
                num_of_signatures += 1
                continue
        if num_of_signatures >= threshold:
            if use_yubikey_for_signing_confirmed:
                if not click.confirm(
                    f"Threshold of {role} keys reached. Do you want to load more {role} keys?"
                ):
                    break
            else:
                # loading from keystore files, couldn't load from all of them, but loaded enough
                break

        # try to load from the inserted YubiKey, without asking the user to insert it
        key_name = get_key_name(role, num_of_signatures, signing_keys_num)
        if _load_and_append_yubikeys(key_name, role, False, True):
            num_of_signatures += 1
            continue

        if prompt_for_yubikey:
            if click.confirm(f"Sign {role} using YubiKey(s)?"):
                use_yubikey_for_signing_confirmed = True
            prompt_for_yubikey = False

        if use_yubikey_for_signing_confirmed:
            if _load_and_append_yubikeys(key_name, role, True, False):
                num_of_signatures += 1
            continue

        if prompt_for_keys and click.confirm(f"Manually enter {role} key?"):
            keys = [signer.public_key for signer in signers_keystore]
            key = key_cmd_prompt(key_name, role, taf_repo, keys, scheme)
            signer = load_signer_from_pem(key)
            signers_keystore.append(key)
            num_of_signatures += 1
        else:
            raise SigningError(f"Cannot load keys of role {role}")

    return signers_keystore, yubikeys


def setup_roles_keys(
    role: Role,
    certs_dir: Optional[Union[Path, str]] = None,
    keystore: Optional[Union[Path, str]] = None,
    yubikeys: Optional[Dict] = None,
    users_yubikeys_details: Optional[Dict[str, UserKeyData]] = None,
    skip_prompt: Optional[bool] = False,
    key_size: int = 2048,
):

    if role.name is None:
        raise SigningError("Cannot set up roles keys. Role name not specified")
    yubikey_keys = []
    keystore_signers = []
    yubikey_signers = []

    yubikey_ids = role.yubikey_ids
    if yubikey_ids is None:
        yubikey_ids = []
    if users_yubikeys_details is None:
        users_yubikeys_details = {}

    is_yubikey = bool(yubikey_ids)

    if is_yubikey:
        yubikey_keys, yubikey_signers = _setup_yubikey_roles_keys(
            yubikey_ids, users_yubikeys_details, yubikeys, role, certs_dir, key_size
        )
    else:
        if keystore is None:
            taf_logger.error("No keystore provided and no default keystore found")
            raise KeyError("No keystore provided and no default keystore found")
        default_params = RoleSetupParams()
        for key_num in range(role.number):
            key_name = get_key_name(role.name, key_num, role.number)
            signer = _setup_keystore_key(
                keystore,
                role.name,
                key_name,
                role.scheme or default_params["scheme"],
                role.length or default_params["length"],
                None,
                skip_prompt=skip_prompt,
            )
            keystore_signers.append(signer)
    return keystore_signers, yubikey_keys, yubikey_signers


def _setup_yubikey_roles_keys(
    yubikey_ids, users_yubikeys_details, yubikeys, role, certs_dir, key_size
):
    loaded_keys_num = 0
    yk_with_public_key = {}
    yubikey_keys = []
    signers = []
    for key_id in yubikey_ids:

        public_key_text = None
        if key_id in users_yubikeys_details:
            public_key_text = users_yubikeys_details[key_id].public
        if public_key_text:
            scheme = users_yubikeys_details[key_id].scheme
            public_key = get_sslib_key_from_value(public_key_text, scheme)
            # Check if the signing key is already loaded
            if not yk.get_key_serial_by_id(key_id):
                yk_with_public_key[key_id] = public_key
            else:
                loaded_keys_num += 1
            yubikey_keys.append(public_key)
        else:
            key_scheme = None
            if key_id in users_yubikeys_details:
                key_scheme = users_yubikeys_details[key_id].scheme
            key_scheme = key_scheme or role.scheme
            public_key, serial_num = _setup_yubikey(
                yubikeys,
                role.name,
                key_id,
                yubikey_keys,
                key_scheme,
                certs_dir,
                key_size,
            )
            loaded_keys_num += 1
            signer = YkSigner(
                public_key, partial(yk.yk_secrets_handler, serial_num=serial_num)
            )
            signers.append(signer)

    if loaded_keys_num < role.threshold:
        print(f"Threshold of role {role.name} is {role.threshold}")
        while loaded_keys_num < role.threshold:
            loaded_keys = []
            for key_id, public_key in yk_with_public_key.items():
                if (
                    key_id in users_yubikeys_details
                    and not users_yubikeys_details[key_id].present
                ):
                    continue
                serial_num = _load_and_verify_yubikey(
                    yubikeys, role.name, key_id, public_key
                )
                if serial_num:
                    loaded_keys_num += 1
                    loaded_keys.append(key_id)
                    signer = YkSigner(
                        public_key,
                        partial(yk.yk_secrets_handler, serial_num=serial_num),
                    )
                    signers.append(signer)
                if loaded_keys_num == role.threshold:
                    break
            if loaded_keys_num < role.threshold:
                if not click.confirm(
                    f"Threshold of signing keys of role {role.name} not reached. Continue?"
                ):
                    raise SigningError("Not enough signing keys")
                for key_id in loaded_keys:
                    yk_with_public_key.pop(key_id)

    return yubikey_keys, signers


def _setup_keystore_key(
    keystore: Optional[Union[Path, str]],
    role_name: str,
    key_name: str,
    scheme: str,
    length: int,
    password: Optional[str],
    skip_prompt: Optional[bool],
) -> CryptoSigner:
    # if keystore exists, load the keys
    generate_new_keys = keystore is None
    signer = None

    def _invalid_key_message(key_name, keystore):
        key_path = Path(keystore, key_name)
        if key_path.is_file():
            print(f"Could not load private key {key_path}")
        else:
            print(f"{key_path} is not a file!")

    if keystore is not None:
        keystore_path = str(Path(keystore).expanduser().resolve())
        while signer is None:
            try:
                signer = load_signer_from_private_keystore(
                    keystore_path,
                    key_name,
                    scheme=scheme,
                    password=password,
                )
            except KeystoreError:
                _invalid_key_message(key_name, keystore)

            if signer is None:
                generate_new_keys = skip_prompt is True or click.confirm(
                    "Generate new keys?"
                )
                if not generate_new_keys:
                    if click.confirm("Reuse existing key?"):
                        reused_key_name = input(
                            "Enter name of an existing keystore file: "
                        )
                        # copy existing private and public keys to the new files
                        Path(keystore, key_name).write_bytes(
                            Path(keystore, reused_key_name).read_bytes()
                        )
                        Path(keystore, key_name + ".pub").write_bytes(
                            Path(keystore, reused_key_name + ".pub").read_bytes()
                        )
                    else:
                        raise KeystoreError(f"Could not load {key_name}")
                else:
                    break
    if generate_new_keys:
        if keystore is not None and (
            skip_prompt or click.confirm("Write keys to keystore files?")
        ):
            if password is None and not skip_prompt:
                password = input(
                    "Enter keystore password and press ENTER (can be left empty)"
                )
            private_pem = generate_and_write_rsa_keypair(
                path=Path(keystore, key_name), key_size=length, password=password
            )
            signer = load_signer_from_pem(private_pem)
        else:
            _, private_pem = generate_rsa_keypair(key_size=length)
            print(f"{role_name} key:\n\n{private_pem.decode()}\n\n")
            signer = load_signer_from_pem(private_pem)

    return signer


def _setup_yubikey(
    yubikeys: Optional[Dict],
    role_name: str,
    key_name: str,
    loaded_keys: List[str],
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    certs_dir: Optional[Union[Path, str]] = None,
    key_size: int = 2048,
) -> Tuple[Dict, str]:
    print(f"Registering keys for {key_name}")
    while True:
        use_existing = click.confirm("Do you want to reuse already set up Yubikey?")
        if not use_existing:
            if not click.confirm(
                "WARNING - this will delete everything from the inserted key. Proceed?"
            ):
                if click.confirm("Cancel?"):
                    raise YubikeyError("Yubikey setup canceled")
                continue
        key, serial_num = yk.yubikey_prompt(
            key_name,
            role_name,
            taf_repo=None,
            registering_new_key=True,
            creating_new_key=not use_existing,
            loaded_yubikeys=yubikeys,
            pin_confirm=True,
            pin_repeat=True,
        )
        if use_existing and key in loaded_keys:
            print("Key already loaded. Please insert a different YubiKey")
        else:
            if not use_existing:
                key = yk.setup_new_yubikey(serial_num, scheme, key_size=key_size)

            if certs_dir is not None:
                yk.export_yk_certificate(certs_dir, key)
            return key, serial_num


def _load_and_verify_yubikey(
    yubikeys: Optional[Dict], role_name: str, key_name: str, public_key
) -> Optional[str]:
    if not click.confirm(f"Sign using {key_name} Yubikey?"):
        return None
    while True:
        yk_public_key, _ = yk.yubikey_prompt(
            key_name,
            role_name,
            taf_repo=None,
            registering_new_key=True,
            creating_new_key=False,
            loaded_yubikeys=yubikeys,
            pin_confirm=True,
            pin_repeat=True,
        )

        if yk_public_key["keyid"] != public_key["keyid"]:
            print("Public key of the inserted key is not equal to the specified one.")
            if not click.confirm("Try again?"):
                return None
        return yk.get_serial_num()



import datetime
from contextlib import contextmanager
from functools import wraps
from collections import defaultdict
from getpass import getpass
from pathlib import Path
from typing import Callable, Dict, Optional

import click
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa, padding

from taf.tuf.keys import get_sslib_key_from_value
from ykman.device import list_all_devices
from yubikit.core.smartcard import SmartCardConnection
from ykman.piv import (
    KEY_TYPE,
    MANAGEMENT_KEY_TYPE,
    SLOT,
    PivSession,
    generate_random_management_key,
)
from yubikit.piv import (
    DEFAULT_MANAGEMENT_KEY,
    PIN_POLICY,
    InvalidPinError,
)

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import InvalidPINError, YubikeyError
from taf.utils import get_pin_for

from securesystemslib.signer._key import SSlibKey

DEFAULT_PIN = "123456"
DEFAULT_PUK = "12345678"
EXPIRATION_INTERVAL = 36500

_yks_data_dict: Dict = defaultdict(dict)


def add_key_id_mapping(serial_num: str, keyid: str) -> None:
    if "ids" not in _yks_data_dict:
        _yks_data_dict["ids"] = defaultdict(dict)
    if keyid not in _yks_data_dict["ids"]:
        _yks_data_dict["ids"][keyid] = serial_num


def add_key_pin(serial_num: str, pin: str) -> None:
    _yks_data_dict[serial_num]["pin"] = pin


def add_key_public_key(serial_num: str, public_key: Dict) -> None:
    _yks_data_dict[serial_num]["public_key"] = public_key


def get_key_pin(serial_num: int) -> Optional[str]:
    if serial_num in _yks_data_dict:
        return _yks_data_dict.get(serial_num, {}).get("pin")
    return None


def get_key_serial_by_id(keyid: str) -> Optional[str]:
    return _yks_data_dict.get("ids", {}).get(keyid)


def get_key_public_key(serial_num: str) -> Optional[Dict]:
    if serial_num in _yks_data_dict:
        return _yks_data_dict.get(serial_num, {}).get("public_key")
    return None


def raise_yubikey_err(msg: Optional[str] = None) -> Callable:
    """Decorator used to catch all errors raised by yubikey-manager and raise
    YubikeyError. We don't need to handle specific cases.
    """

    def wrapper(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except YubikeyError:
                raise
            except Exception as e:
                err_msg = (
                    f"{msg} Reason: ({type(e).__name__}) {str(e)}" if msg else str(e)
                )
                raise YubikeyError(err_msg) from e

        return decorator

    return wrapper


@contextmanager
def _yk_piv_ctrl(serial=None):
    """Context manager to open connection and instantiate Piv Session.

    Args:
        - serial (str): Match Yubikey's serial multiple keys are inserted

    Returns:
        - ykman.piv.PivSession

    Raises:
        - YubikeyError
    """
    # If pub_key_pem is given, iterate all devices, read x509 certs and try to match
    # public keys.
    for dev, info in list_all_devices():
        if serial is None or info.serial == serial:
            with dev.open_connection(SmartCardConnection) as connection:
                session = PivSession(connection)
                yield session, info.serial
        else:
            pass


def is_inserted():
    """Checks if YubiKey is inserted.

    Args:
        None

    Returns:
        True if at least one Yubikey is inserted (bool)

    Raises:
        - YubikeyError
    """
    return len(list(list_all_devices())) > 0


@raise_yubikey_err()
def is_valid_pin(pin):
    """Checks if given pin is valid.

    Args:
        pin(str): Yubikey piv PIN

    Returns:
        tuple: True if PIN is valid, otherwise False, number of PIN retries

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        try:
            ctrl.verify_pin(pin)
            return True, None  # ctrl.get_pin_tries() fails if PIN is valid
        except InvalidPinError:
            return False, ctrl.get_pin_attempts()


@raise_yubikey_err("Cannot get serial number.")
def get_serial_num():
    """Get Yubikey serial number.

    Args:
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Yubikey serial number

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (_, serial):
        return serial


@raise_yubikey_err("Cannot export x509 certificate.")
def export_piv_x509(cert_format=serialization.Encoding.PEM):
    """Exports YubiKey's piv slot x509.

    Args:
        - cert_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV x509 certificate in a given format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        x509 = ctrl.get_certificate(SLOT.SIGNATURE)
        return x509.public_bytes(encoding=cert_format)


@raise_yubikey_err("Cannot export public key.")
def export_piv_pub_key(pub_key_format=serialization.Encoding.PEM):
    """Exports YubiKey's piv slot public key.

    Args:
        - pub_key_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV public key in a given format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        try:
            x509_cert = ctrl.get_certificate(SLOT.SIGNATURE)
            public_key = x509_cert.public_key()
            return public_key.public_bytes(
                encoding=pub_key_format,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        except Exception as e:
            raise YubikeyError(f"Failed to export public key: {str(e)}") from e


@raise_yubikey_err("Cannot export yk certificate.")
def export_yk_certificate(certs_dir, key: SSlibKey):
    if certs_dir is None:
        certs_dir = Path.home()
    else:
        certs_dir = Path(certs_dir)
    certs_dir.mkdir(parents=True, exist_ok=True)
    cert_path = certs_dir / f"{key.keyid}.cert"
    print(f"Exporting certificate to {cert_path}")
    with open(cert_path, "wb") as f:
        f.write(export_piv_x509())


@raise_yubikey_err("Cannot get public key in TUF format.")
def get_piv_public_key_tuf(scheme=DEFAULT_RSA_SIGNATURE_SCHEME) -> SSlibKey:
    """Return public key from a Yubikey in TUF's RSAKEY_SCHEMA format.

    Args:
        - scheme(str): Rsa signature scheme (default is rsa-pkcs1v15-sha256)
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        A dictionary containing the RSA keys and other identifying information
        from inserted smart card.
        Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.

    Raises:
        - YubikeyError
    """
    pub_key_pem = export_piv_pub_key().decode("utf-8")
    return get_sslib_key_from_value(pub_key_pem, scheme)


@raise_yubikey_err("Cannot sign data.")
def sign_piv_rsa_pkcs1v15(data, pin):
    """Sign data with key from YubiKey's piv slot.

    Args:
        - data(bytes): Data to be signed
        - pin(str): Pin for piv slot login.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Signature (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        ctrl.verify_pin(pin)
        return ctrl.sign(
            SLOT.SIGNATURE, KEY_TYPE.RSA2048, data, hashes.SHA256(), padding.PKCS1v15()
        )


@raise_yubikey_err("Cannot setup Yubikey.")
def setup(
    pin,
    cert_cn,
    cert_exp_days=365,
    pin_retries=10,
    private_key_pem=None,
    mgm_key=generate_random_management_key(MANAGEMENT_KEY_TYPE.TDES),
    key_size=2048,
):
    """Use to setup inserted Yubikey, with following steps (order is important):
      - reset to factory settings
      - set management key
      - generate key(RSA2048) or import given one
      - generate and import self-signed certificate(X509)
      - set pin retries
      - set pin
      - set puk(same as pin)

    Args:
        - cert_cn(str): x509 common name
        - cert_exp_days(int): x509 expiration (in days from now)
        - pin_retries(int): Number of retries for PIN
        - private_key_pem(str): Private key in PEM format. If given, it will be
                                imported to Yubikey.
        - mgm_key(bytes): New management key

    Returns:
        PIV public key in PEM format (bytes)

    Raises:
        - YubikeyError
    """

    with _yk_piv_ctrl() as (ctrl, _):
        # Factory reset and set PINs
        ctrl.reset()

        ctrl.authenticate(MANAGEMENT_KEY_TYPE.TDES, DEFAULT_MANAGEMENT_KEY)
        ctrl.set_management_key(MANAGEMENT_KEY_TYPE.TDES, mgm_key)

        # Generate RSA2048
        if private_key_pem is None:
            private_key = rsa.generate_private_key(65537, key_size, default_backend())
            pub_key = private_key.public_key()
        else:
            try:
                private_key = load_pem_private_key(
                    private_key_pem, None, default_backend()
                )
            except TypeError:
                pem_pwd = getpass("Enter pem file password:\n")
                if pem_pwd:
                    pem_pwd = pem_pwd.encode()
                private_key = load_pem_private_key(
                    private_key_pem, pem_pwd, default_backend()
                )

        ctrl.put_key(SLOT.SIGNATURE, private_key, PIN_POLICY.ALWAYS)
        pub_key = private_key.public_key()
        ctrl.authenticate(MANAGEMENT_KEY_TYPE.TDES, mgm_key)
        ctrl.verify_pin(DEFAULT_PIN)

        now = datetime.datetime.now()
        valid_to = now + datetime.timedelta(days=cert_exp_days)

        name = x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME, cert_cn)])
        # Generate and import certificate
        cert = (
            x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(pub_key)
            .serial_number(x509.random_serial_number())
            .not_valid_before(now)
            .not_valid_after(valid_to)
            .sign(private_key, hashes.SHA256(), default_backend())
        )

        ctrl.put_certificate(SLOT.SIGNATURE, cert)

        ctrl.set_pin_attempts(pin_attempts=pin_retries, puk_attempts=pin_retries)
        ctrl.change_pin(DEFAULT_PIN, pin)
        ctrl.change_puk(DEFAULT_PUK, pin)

    return pub_key.public_bytes(
        serialization.Encoding.PEM, serialization.PublicFormat.SubjectPublicKeyInfo
    )


def setup_new_yubikey(
    serial_num, scheme=DEFAULT_RSA_SIGNATURE_SCHEME, key_size=2048
) -> SSlibKey:
    pin = get_key_pin(serial_num)
    cert_cn = input("Enter key holder's name: ")
    print("Generating key, please wait...")
    pub_key_pem = setup(
        pin, cert_cn, cert_exp_days=EXPIRATION_INTERVAL, key_size=key_size
    ).decode("utf-8")
    scheme = DEFAULT_RSA_SIGNATURE_SCHEME
    key = get_sslib_key_from_value(pub_key_pem, scheme)
    return key


def get_and_validate_pin(key_name, pin_confirm=True, pin_repeat=True):
    valid_pin = False
    while not valid_pin:
        pin = get_pin_for(key_name, pin_confirm, pin_repeat)
        valid_pin, retries = is_valid_pin(pin)
        if not valid_pin and not retries:
            raise InvalidPINError("No retries left. YubiKey locked.")
        if not valid_pin:
            if not click.confirm(
                f"Incorrect PIN. Do you want to try again? {retries} retires left."
            ):
                raise InvalidPINError("PIN input cancelled")
    return pin


def yubikey_prompt(
    key_name,
    role=None,
    taf_repo=None,
    registering_new_key=False,
    creating_new_key=False,
    loaded_yubikeys=None,
    pin_confirm=True,
    pin_repeat=True,
    prompt_message=None,
    retry_on_failure=True,
    hide_already_loaded_message=False,
):
    def _read_and_check_yubikey(
        key_name,
        role,
        taf_repo,
        registering_new_key,
        creating_new_key,
        loaded_yubikeys,
        pin_confirm,
        pin_repeat,
        prompt_message,
        retrying,
    ):

        if retrying:
            if prompt_message is None:
                prompt_message = f"Please insert {key_name} YubiKey and press ENTER"
            getpass(prompt_message)
        # make sure that YubiKey is inserted
        try:
            serial_num = get_serial_num()
        except Exception:
            print("YubiKey not inserted")
            return False, None, None

        # check if this key is already loaded as the provided role's key (we can use the same key
        # to sign different metadata)
        if (
            loaded_yubikeys is not None
            and serial_num in loaded_yubikeys
            and role in loaded_yubikeys[serial_num]
        ):
            if not hide_already_loaded_message:
                print("Key already loaded")
            return False, None, None

        # read the public key, unless a new key needs to be generated on the yubikey
        public_key = get_piv_public_key_tuf() if not creating_new_key else None
        # check if this yubikey is can be used for signing the provided role's metadata
        # if the key was already registered as that role's key
        if not registering_new_key and role is not None and taf_repo is not None:
            if not taf_repo.is_valid_metadata_yubikey(role, public_key):
                print(f"The inserted YubiKey is not a valid {role} key")
                return False, None, None

        if get_key_pin(serial_num) is None:
            if creating_new_key:
                pin = get_pin_for(key_name, pin_confirm, pin_repeat)
            else:
                pin = get_and_validate_pin(key_name, pin_confirm, pin_repeat)
            add_key_pin(serial_num, pin)

        if get_key_public_key(serial_num) is None and public_key is not None:
            add_key_public_key(serial_num, public_key)

        # when reusing the same yubikey, public key will already be in the public keys dictionary
        # but the key name still needs to be added to the key id mapping dictionary
        add_key_id_mapping(serial_num, key_name)

        if role is not None:
            if loaded_yubikeys is None:
                loaded_yubikeys = {serial_num: [role]}
            else:
                loaded_yubikeys.setdefault(serial_num, []).append(role)

        return True, public_key, serial_num

    retry_counter = 0
    while True:
        success, key, serial_num = _read_and_check_yubikey(
            key_name,
            role,
            taf_repo,
            registering_new_key,
            creating_new_key,
            loaded_yubikeys,
            pin_confirm,
            pin_repeat,
            prompt_message,
            retrying=retry_counter > 0,
        )
        if not success and not retry_on_failure:
            return None, None
        if success:
            return key, serial_num
        retry_counter += 1


def yk_secrets_handler(prompt, serial_num):
    if prompt == "pin":
        return get_key_pin(serial_num)
    raise YubikeyError(f"Invalid prompt {prompt}")



import click
import json
from taf.api.yubikey import (
    export_yk_certificate,
    export_yk_public_pem,
    get_yk_roles,
    setup_signing_yubikey,
    setup_test_yubikey,
)
from taf.exceptions import YubikeyError
from taf.repository_utils import find_valid_repository
from taf.tools.cli import catch_cli_exception


def check_pin_command():
    @click.command(help="Checks if the specified pin is valid")
    @click.argument("pin")
    def check_pin(pin):
        try:
            from taf.yubikey import is_valid_pin

            valid, retries = is_valid_pin(pin)
            inserted = True
        except YubikeyError:
            valid = False
            inserted = False
            retries = None
        print(json.dumps({"pin": valid, "retries": retries, "inserted": inserted}))

    return check_pin


def export_pub_key_command():
    @click.command(
        help="Export the inserted Yubikey's public key and save it to the specified location."
    )
    @click.option(
        "--output",
        help="File to which the exported public key will be written. The result will be written to the console if path is not specified",
    )
    def export_pub_key(output):
        export_yk_public_pem(output)

    return export_pub_key


def get_roles_command():
    @click.command(
        help="Export the inserted Yubikey's public key and save it to the specified location."
    )
    @catch_cli_exception(handle=YubikeyError, print_error=True)
    @click.option(
        "--path",
        default=".",
        help="Authentication repository's location. If not specified, set to the current directory",
    )
    def get_roles(path):
        path = find_valid_repository(path)
        roles_with_paths = get_yk_roles(path)
        for role, paths in roles_with_paths.items():
            print(f"\n{role}")
            for path in paths:
                print(f"\n -{path}")

    return get_roles


def export_certificate_command():
    @click.command(
        help="Export the inserted Yubikey's public key and save it to the specified location."
    )
    @click.option(
        "--output",
        help="File to which the exported certificate key will be written. The result will be written to the user's home directory by default",
    )
    def export_certificate(output):
        export_yk_certificate(output)

    return export_certificate


def setup_signing_key_command():
    @click.command(
        help="""Generate a new key on the yubikey and set the pin. Export the generated certificate
        to the specified directory.
        WARNING - this will delete everything from the inserted key."""
    )
    @click.option(
        "--certs-dir",
        help="Path of the directory where the exported certificate will be saved. Set to the user home directory by default",
    )
    def setup_signing_key(certs_dir):
        setup_signing_yubikey(certs_dir, key_size=2048)

    return setup_signing_key


def setup_test_key_command():
    @click.command(
        help="""Copies the specified key onto the inserted YubiKey
        WARNING - this will reset the inserted key."""
    )
    @click.argument("key-path")
    def setup_test_key(key_path):
        setup_test_yubikey(key_path)

    return setup_test_key


def attach_to_group(group):
    group.add_command(check_pin_command(), name="check-pin")
    group.add_command(export_pub_key_command(), name="export-pub-key")
    group.add_command(get_roles_command(), name="get-roles")
    group.add_command(export_certificate_command(), name="export-certificate")
    group.add_command(setup_signing_key_command(), name="setup-signing-key")
    group.add_command(setup_test_key_command(), name="setup-test-key")



"""
This module contains utility functions that cause circular import errors in utils.py.
"""

from pathlib import Path
from taf.exceptions import GitError, InvalidRepositoryError
from taf.auth_repo import AuthenticationRepository
from taf.git import GitRepository
from taf.utils import taf_logger


def find_valid_repository(path: Path) -> Path:
    """
    Find a valid authentication repository starting from the given path and traversing subdirectories and parent directories if needed.
    """

    def try_load_repository(repo_path: Path) -> bool:
        try:
            auth_repo = AuthenticationRepository(path=repo_path)
            if auth_repo.get_metadata("root") is not None:
                taf_logger.info(
                    f"Loaded valid authentication repository from {repo_path}"
                )
                return True
            return False
        except GitError:
            return False

    path = Path(path).resolve()
    if not path.is_dir():
        raise InvalidRepositoryError(f"Directory {path} does not exist")
    # First, try to load the repository from the given path
    if try_load_repository(path):
        # find the git repository's root
        current_path = path
        while not GitRepository(path=current_path).is_git_repository_root:
            current_path = current_path.parent
        return current_path

    # Search subdirectories
    taf_logger.debug(
        f"Current directory {path} is not a valid authentication repository. Searching subdirectories..."
    )
    for subdir in path.iterdir():
        if subdir.is_dir() and try_load_repository(subdir):
            return subdir

    raise InvalidRepositoryError(
        f"Could not find a valid authentication repository in {path} or any of its subdirectories or parent directories."
    )



"""
The general idea of the updater is the following:
- We have a git repository which contains the metadata files. These metadata files
are in the 'metadata' directory
- Clients have a clone of that repository on their local machine and want to update it
- We don't want to simply pull the updates. We want to verify that the new commits
(committed after the most recent one in the client's local repository)
- For each of the new commits, we want to check if all metadata is valid. The set of
metadata should be valid as a whole at that revision. Not only do we want to make sure
that a metadata which is supposed to be changed was indeed updated and is valid, but
also to make sure that if a metadata file should not be updated, it remained the same.
- We also want to make sure that all targets metadata is valid (including the delegated roles)
- We do not want to simply update the metadata to the latest version, without skipping
these checks. We want to check each commit, not just the last one.
- If we are checking a commit which is not the latest one, we do not want to report an error
if the metadata expired. We want to make sure that that was valid at the time when the
metadata was committed.
- We can rely on the TUF's way of handling metadata, by using the current and previous
directories. We just want to automatically create and update them. They should not
remain on the client's machine.
- We do not want to modify TUF's updater to much, but still need to get around the fact
that TUF skips mirrors which do not have valid and/or current metadata files. Also, we
do not simply want to find the latest metadata, we want to validate everything in-between.
That is why the idea is to call refresh multiple times, until the last commit is reached.
The 'GitUpdater' updater is designed in such a way that for each new call it
loads data from a most recent commit.
"""
import copy
from logging import ERROR

from typing import Dict, Tuple, Any
from attr import define, field
from logdecorator import log_on_error
from taf.auth_repo import AuthenticationRepository
from taf.git import GitRepository
from taf.updater.types.update import OperationType, UpdateType
from taf.updater.updater_pipeline import (
    AuthenticationRepositoryUpdatePipeline,
)

from pathlib import Path
from taf.log import taf_logger
import taf.repositoriesdb as repositoriesdb
from taf.utils import is_non_empty_directory, timed_run
import taf.settings as settings
from taf.exceptions import (
    ScriptExecutionError,
    UpdateFailedError,
    ValidationFailedError,
)

from taf.updater.lifecycle_handlers import (
    handle_repo_event,
    handle_update_event,
    Event,
)
from cattr import unstructure
import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
from taf.updater.types.update import Update


def _check_update_status(repos_update_data: Dict[str, Any]) -> Tuple[Event, str]:
    # helper function to set update status of update handler based on repo status.
    # if repo handlers event status changed,
    # change the update handler status
    update_status = Event.UNCHANGED
    errors = ""

    for auth_repo_name in repos_update_data:
        repo_update_data = repos_update_data[auth_repo_name]
        repo_update_status = repo_update_data["update_status"]
        if update_status != repo_update_status and update_status != Event.FAILED:
            # if one failed, then failed
            # else if one changed, then changed
            # else unchanged
            update_status = repo_update_status
        repo_error = repo_update_data["error"]
        if repo_error is not None:
            errors += str(repo_error)

    return update_status, errors


def _execute_repo_handlers(
    update_status,
    auth_repo,
    scripts_root_dir,
    commits_data,
    error,
    warnings,
    targets_data,
    transient_data,
):
    try:
        transient = handle_repo_event(
            update_status,
            None,
            auth_repo.library_dir,
            scripts_root_dir,
            auth_repo,
            commits_data,
            error,
            warnings,
            targets_data,
        )
        if transient_data is not None:
            transient_data.update(transient)
    except ScriptExecutionError as e:
        if settings.development_mode:
            _reset_to_commits_before_pull(auth_repo, commits_data, targets_data)
            error = e


def _reset_to_commits_before_pull(auth_repo, commits_data, targets_data):
    taf_logger.info(
        "In development mode. Resetting repositories to commits before pull after handler failure"
    )

    def _reset_repository(repo, commits_data):
        before_pull = commits_data.before_pull
        after_pull = commits_data.after_pull
        if before_pull == after_pull:
            return
        repo.reset_to_commit(before_pull, hard=True)

    auth_repo.checkout_branch(auth_repo.default_branch)
    _reset_repository(auth_repo, commits_data)
    auth_repo.set_last_validated_commit(commits_data.before_pull)

    for repo_name, repo_data in targets_data.items():
        repo = repositoriesdb.get_repository(auth_repo, repo_name)
        for branch, branch_data in repo_data["commits"].items():
            repo.checkout_branch(branch)
            _reset_repository(repo, branch_data)


@define
class UpdateConfig:
    operation: OperationType = field(converter=OperationType)
    remote_url: str = field(
        metadata={"docs": "Remote URL of the remote authentication repository"},
        default=None,
    )
    path: Path = field(
        default=None,
        converter=lambda p: Path(p).resolve() if p else None,
        metadata={"docs": "Client's authentication repository's full path"},
    )
    library_dir: Path = field(
        default=None,
        metadata={
            "docs": "Directory where client's target repositories are located. Optional."
        },
    )
    update_from_filesystem: bool = field(
        default=False,
        metadata={
            "docs": "A flag indicating if the URL is actually a file system path. Optional."
        },
    )
    expected_repo_type: UpdateType = field(
        default=UpdateType.EITHER,
        metadata={
            "docs": "Indicates if the repository is a test, official, or any type. Optional."
        },
    )
    target_repo_classes: object = field(
        default=None,
        metadata={
            "docs": "A class or dictionary used for instantiating target repositories. Optional."
        },
    )
    target_factory: object = field(
        default=None,
        metadata={
            "docs": "A git repositories factory used for instantiating target repositories. Optional."
        },
    )
    only_validate: bool = field(
        default=False,
        metadata={
            "docs": "Specifies if repositories should only be validated without being updated. Optional."
        },
    )
    validate_from_commit: str = field(
        default=None,
        metadata={"docs": "Commit from which validation should start. Optional."},
    )
    out_of_band_authentication: str = field(
        default=None,
        metadata={"docs": "Out-of-band authentication commit's SHA. Optional."},
    )
    scripts_root_dir: Path = field(
        default=None,
        metadata={
            "docs": "Local directory for script testing, not in the authentication repository. Optional."
        },
    )
    checkout: bool = field(
        default=True,
        metadata={
            "docs": "Whether to checkout last validated commits after update. Optional."
        },
    )
    clone_urls: list = field(
        default=None,
        metadata={"docs": "List of URLs to clone repositories from. Optional."},
    )
    excluded_target_globs: list = field(
        default=None,
        metadata={
            "docs": "Globs specifying target repositories to exclude from validation and update. Optional."
        },
    )
    strict: bool = field(
        default=False,
        metadata={"docs": "Whether update fails if a warning is raised. Optional."},
    )
    bare: bool = field(
        default=False,
        metadata={
            "docs": "Whether to clone repositories as bare repositories. If set to true, all repositories will be cloned as bare repositories. Optional."
        },
    )
    force: bool = field(
        default=False,
        metadata={
            "docs": "Whether to force update repositories. If set to true, all repositories will be forcefully updated."
        },
    )
    no_deps: bool = field(
        default=False,
        metadata={"docs": "Specifies whether or not to update dependencies. Optional."},
    )
    no_targets: bool = field(
        default=False,
        metadata={
            "docs": "Flag to skip target repositiory validation and validate only authentication repos. Optional."
        },
    )
    no_upstream: bool = field(
        default=True,
        metadata={
            "docs": "Flag to skip comparison with remote repositories upstream. Optional."
        },
    )

    def __attrs_post_init__(self):
        if self.operation == OperationType.CLONE:
            if self.library_dir is None:
                if self.path:
                    self.library_dir = self.path.parent.parent
                else:
                    self.library_dir = Path(".").resolve()

        if self.operation == OperationType.UPDATE:
            if self.path is None:
                self.path = Path(".").resolve()
            if self.library_dir is None:
                self.library_dir = self.path.parent.parent


@log_on_error(
    ERROR,
    "{e}",
    logger=taf_logger,
    on_exceptions=UpdateFailedError,
    reraise=True,
)
@timed_run("Cloning repositories")
def clone_repository(config: UpdateConfig):
    """
    Validate and clone an authentication repository and its target repositories, as well
    as its dependencies (linked authentication repositories and their targets) recursively.
    Arguments:
        config: UpdateConfig instance containing all configurations.

    Side Effects:
        If only_validate is not set to True, updates authentication repository (pulls new changes) and its target
        repositories and dependencies

    Returns:
        None
    """
    settings.strict = config.strict

    if config.remote_url is None:
        raise UpdateFailedError(
            "Remote URL has to be specified when cloning repositories"
        )

    if config.path and is_non_empty_directory(config.path):
        raise UpdateFailedError(
            f"Destination path {config.path} already exists and is not an empty directory. Run 'taf repo update' to update it."
        )

    config.operation = OperationType.CLONE
    return _update_or_clone_repository(config)


@log_on_error(
    ERROR,
    "{e}",
    logger=taf_logger,
    on_exceptions=UpdateFailedError,
    reraise=True,
)
@timed_run("Updating repository")
def update_repository(config: UpdateConfig):
    """
    Validate and update an authentication repository and its target repositories, as well
    as its dependencies (linked authentication repositories and their targets) recursively.

    Arguments:
        config: UpdateConfig instance containing all configurations.

    Side Effects:
        If only_validate is not set to True, updates authentication repository (pulls new changes) and its target
        repositories and dependencies

    Returns:
        None
    """
    settings.strict = config.strict

    # if path is not specified, name should be read from info.json
    # which is available after the remote repository is cloned and validated

    auth_repo = GitRepository(path=config.path)
    if not config.path.is_dir() or not auth_repo.is_git_repository:
        raise UpdateFailedError(
            f"{config.path} is not a Git repository. Run 'taf repo clone' instead"
        )

    taf_logger.info(f"Updating repository {auth_repo.name}")

    if config.remote_url is None:
        config.remote_url = auth_repo.get_remote_url()
        if config.remote_url is None:
            raise UpdateFailedError("URL cannot be determined. Please specify it")

    if auth_repo.is_bare_repository:
        # Handle updates for bare repositories
        config.bare = True
    return _update_or_clone_repository(config)


def _update_or_clone_repository(config: UpdateConfig):
    repos_update_data: Dict = {}
    transient_data: Dict = {}
    root_error = None
    auth_repo_name = None
    try:
        updater_pipeline = AuthenticationRepositoryUpdatePipeline(config)
        updater_pipeline.run()
        update_output = updater_pipeline.output
        auth_repo_name = update_output.auth_repo_name
        _process_repo_update(
            update_config=config,
            update_output=update_output,
            repos_update_data=repos_update_data,
            transient_data=transient_data,
        )
        if repos_update_data[auth_repo_name].get("error"):
            raise repos_update_data[auth_repo_name]["error"]

    except Exception as e:
        root_error = UpdateFailedError(
            f"Update of {auth_repo_name or 'repository'} failed due to error: {e}"
        )

    update_data = Update()

    if auth_repo_name is None or auth_repo_name not in repos_update_data:
        # this must mean that an error occurred
        if root_error is not None:
            raise root_error
        else:
            raise UpdateFailedError(f"Update of {auth_repo_name} failed")

    # after all repositories have been updated
    # update information is in repos_update_data
    root_auth_repo = repos_update_data[auth_repo_name]["auth_repo"]

    update_status, errors = _check_update_status(repos_update_data)
    update_transient_data = _update_transient_data(transient_data, repos_update_data)

    update_data = handle_update_event(
        update_status,
        update_transient_data,
        root_auth_repo.library_dir,
        config.scripts_root_dir,
        repos_update_data,
        errors,
        root_auth_repo,
    )
    if repos_update_data[auth_repo_name].get("warnings"):
        taf_logger.warning(repos_update_data[auth_repo_name].get("warnings"))

    log_repo_updates(update_data)

    if root_error:
        raise root_error
    return unstructure(update_data)


def _process_repo_update(
    update_config,
    update_output,
    visited=None,
    repos_update_data=None,
    transient_data=None,
):
    """
    Arguments:
        update_config: update configuration object containing all relevant information, like remote url, operation type, local paths
        visited (optional): Authentication repositories which were already processed
        repos_update_data (optional): update status, commits data, targets data of the repository which was updated
        transient_data (optinal): data passed from one lifecycle handler to the next one

    """

    if visited is None:
        visited = []
    # if there is a recursive dependency
    if update_config.remote_url in visited:
        return
    visited.append(update_config.remote_url)
    # at the moment, we assume that the initial commit is valid and that it contains at least root.json
    update_status = update_output.event
    auth_repo = update_output.users_auth_repo
    commits_data = update_output.commits_data
    error = update_output.error
    warnings = update_output.warnings
    targets_data = (update_output.targets_data,)

    # if auth_repo doesn't exist, means that either clients-auth-path isn't provided,
    # or info.json is missing from protected
    if auth_repo is None:
        raise error

    # if commits_data is empty, do not attempt to load the dependencies
    # that can happen in the repository didn't exists, but could not be validated
    # and was therefore deleted
    # or if the last validated commit is not equal to the top commit, meaning that
    # the repository was updated without using the updater
    # this second case could be reworked to return the state as of the last validated commit
    # but treat the repository as invalid for now
    commits = []

    if commits_data.after_pull is not None:
        if commits_data.before_pull is not None:
            commits = [commits_data.before_pull]
        commits.extend(commits_data.new)

    if commits_data.after_pull is not None and not update_config.no_deps:

        if update_status != Event.FAILED:
            # for now, just take the newest commit and do not worry about updated definitions
            # latest_commit = commits[-1::]
            repositoriesdb.load_dependencies(
                auth_repo,
                library_dir=update_config.library_dir,
                commits=commits,
            )

            # load the repositories from dependencies.json and update these repositories
            child_auth_repos = repositoriesdb.get_deduplicated_auth_repositories(
                auth_repo, commits
            ).values()
            outputs, errors = _update_dependencies(update_config, child_auth_repos)
            if len(errors):
                errors = "\n".join(errors)
                taf_logger.error(
                    "Update of {} failed. One or more referenced authentication repositories could not be validated:\n {}",
                    auth_repo.name,
                    errors,
                )
                error = UpdateFailedError(
                    f"Update of {auth_repo.name} failed. One or more referenced authentication repositories could not be validated:\n {errors}"
                )
            for output in outputs:
                child_update_status = output.event
                if child_update_status == Event.FAILED:
                    update_status = Event.FAILED
                repo = output.users_auth_repo
                child_config = copy.copy(update_config)
                child_config.remote_url = repo.urls[0]
                child_config.clone_urls = repo.urls
                child_config.out_of_band_authentication = (
                    repo.out_of_band_authentication
                )
                child_config.path = repo.path
                _process_repo_update(
                    child_config, output, visited, repos_update_data, transient_data
                )

        # do not call the handlers if only validating the repositories
        # if a handler fails and we are in the development mode, revert the update
        # so that it's easy to try again after fixing the handler
        if (
            not update_config.only_validate
            and not update_config.excluded_target_globs
            and not update_config.no_targets
        ):
            _execute_repo_handlers(
                update_status,
                auth_repo,
                update_config.scripts_root_dir,
                commits_data,
                error,
                warnings,
                targets_data,
                transient_data,
            )

    if repos_update_data is not None:
        repos_update_data[auth_repo.name] = {
            "auth_repo": auth_repo,
            "update_status": update_status,
            "commits_data": commits_data,
            "error": error,
            "warnings": warnings,
            "targets_data": targets_data,
        }

    repositoriesdb.clear_repositories_db()


def log_repo_updates(update_data: Update):
    """
    Log the status of the repositories after updating them.
    """
    update_output_dict = {
        repo_name: {
            "changed": repo_info["changed"],
            "event": repo_info["event"],
            **({"warning": repo_info["warnings"]} if "warnings" in repo_info else {}),
            **({"error": repo_info["error_msg"]} if repo_info["error_msg"] else {}),
        }
        for repo_name, repo_info in update_data.auth_repos.items()
    }
    changes_or_errors = False

    for repo_name, details in update_output_dict.items():
        changed = details.get("changed", False)
        event_operation = details["event"]
        error_message = details.get("error")

        if changed or error_message:
            changes_or_errors = True
            message = f"Repository: {repo_name}\n"
            message += f"  Change status: {'changed' if changed else 'no changes'}\n"
            message += f"  Event: {event_operation}"

            if error_message:
                message += f"\n  Error: {error_message}\n"

            taf_logger.log("NOTICE", message)

    if not changes_or_errors:
        taf_logger.log(
            "NOTICE", "All repositories are up-to-date with no changes or errors."
        )


def _update_dependencies(update_config, child_auth_repos):

    # for now, just take the newest commit and do not worry about updated definitions
    # latest_commit = commits[-1::]
    outputs = []
    errors = []

    def _update_child_repo(updater_pipeline):
        try:
            updater_pipeline.run()
            output = updater_pipeline.output
            error = output.error
            return output, error
        except Exception as e:
            return None, e

    with ThreadPoolExecutor() as executor:
        futures = {}
        for repo in child_auth_repos:
            child_config = copy.copy(update_config)
            child_config.operation = (
                OperationType.UPDATE if repo.is_git_repository else OperationType.CLONE
            )
            child_config.remote_url = repo.urls[0]
            child_config.clone_urls = repo.urls
            child_config.out_of_band_authentication = repo.out_of_band_authentication
            child_config.path = repo.path
            pipeline = AuthenticationRepositoryUpdatePipeline(child_config)
            future = executor.submit(_update_child_repo, pipeline)
            futures[future] = repo

        for future in concurrent.futures.as_completed(futures):
            output, error = future.result()
            if error:
                errors.append(str(error))
            if output:
                outputs.append(output)
    return outputs, errors


def _update_transient_data(
    transient_data, repos_update_data: Dict[str, str]
) -> Dict[str, Any]:
    update_transient_data = {}
    for auth_repo_name in repos_update_data:
        if auth_repo_name in transient_data:
            update_transient_data[auth_repo_name] = transient_data[auth_repo_name]
    return update_transient_data


@timed_run("Validating repository")
def validate_repository(
    auth_path,
    library_dir=None,
    validate_from_commit=None,
    excluded_target_globs=None,
    strict=False,
    bare=False,
    no_targets=False,
    no_deps=False,
):
    settings.strict = strict

    auth_path = Path(auth_path).resolve()

    if library_dir is None:
        library_dir = auth_path.parent.parent
    else:
        library_dir = Path(library_dir).resolve()

    auth_repo = AuthenticationRepository(path=auth_path)
    expected_repo_type = (
        UpdateType.TEST if auth_repo.is_test_repo else UpdateType.OFFICIAL
    )

    auth_repo_name = None

    try:
        config = UpdateConfig(
            operation=OperationType.UPDATE,
            remote_url=str(auth_path),
            path=auth_path,
            library_dir=library_dir,
            validate_from_commit=validate_from_commit,
            excluded_target_globs=excluded_target_globs,
            strict=strict,
            bare=bare,
            no_targets=no_targets,
            no_deps=no_deps,
            expected_repo_type=expected_repo_type,
            update_from_filesystem=True,
            only_validate=True,
        )
        updater_pipeline = AuthenticationRepositoryUpdatePipeline(config)
        updater_pipeline.run()
        update_output = updater_pipeline.output
        auth_repo_name = update_output.auth_repo_name
        if update_output.error:
            raise update_output.error

        _process_repo_update(
            config,
            update_output,
        )
    except Exception as e:
        raise ValidationFailedError(
            f"Validation of repository {auth_repo_name or ''} failed due to error: {e}"
        )
    settings.last_validated_commit = {}



"""Add platform-dependent libraries to the path.
"""
import sys


class YubikeyMissingLibrary:
    """If `yubikey-manager` is not installed and we try to use any function from `taf.yubikey`
    module, we will log appropriate error message and exit with code 1.
    """

    ERR_MSG = '"yubikey-manager" is not installed. Run "pip install taf[yubikey]" to install it.'

    def __getattr__(self, name):
        from taf.log import taf_logger

        taf_logger.warning(YubikeyMissingLibrary.ERR_MSG)
        sys.exit(1)



from logging import INFO
from typing import Optional, Union
from logdecorator import log_on_start, log_on_end
from pathlib import Path
from taf.models.types import RolesKeysData
from taf.api.utils._conf import find_taf_directory

from taf.api.roles import _initialize_roles_and_keystore
from taf.keys import get_key_name
from taf.log import taf_logger
from taf.models.types import RolesIterator
from taf.models.converter import from_dict
from taf.exceptions import KeystoreError
from taf.tuf.keys import generate_and_write_rsa_keypair, generate_rsa_keypair


@log_on_start(INFO, "Generating '{key_path:s}'", logger=taf_logger)
@log_on_end(INFO, "Generated '{key_path:s}", logger=taf_logger)
def _generate_rsa_key(key_path: str, password: str, bits: Optional[int] = None) -> None:
    """
    Generate public and private key

    Arguments:
        key_path (optional): The path to write the private key to.
        password (optional): An encryption password.
        bits (optional): The number of bits of the generated RSA key.

    Raises:
        UnsupportedLibraryError: pyca/cryptography is not available.
        FormatError: Arguments are malformed.
        StorageError: Key files cannot be written.

    Side Effects:
        Writes key files to disk.
        Overwrites files if they already exist.

    Returns:
        None
    """
    try:
        generate_and_write_rsa_keypair(path=key_path, key_size=bits, password=password)
        taf_logger.log("NOTICE", f"Generated key {key_path}")
    except Exception:
        taf_logger.error(f"An error occurred while generating rsa key {key_path}")
        raise KeystoreError(f"An error occurred while generating rsa key {key_path}")


def generate_keys(
    keystore: Optional[Union[str, Path]], roles_key_infos: Optional[str]
) -> None:
    """
    Generate public and private keys and writes them to disk. Names of keys correspond to names
    of TUF roles. If more than one key should be generated per role, a counter is appended
    to the role's name. E.g. root1, root2, root3 etc.

    Arguments:
        keystore: Location where the generated files should be saved
        roles_key_infos: Path to a json file which contains information about repository's roles and keys.
    Raises:
        UnsupportedLibraryError: pyca/cryptography is not available.
        FormatError: One or more keys not properly specified
        StorageError: Key files cannot be written.

    Side Effects:
        Writes key files to disk.
        Overwrites files if they already exist.

    Returns:
        None

    Raises:
        KeystoreError if an error occurs while initializing the keystore directory or generating a key
    """
    # TODO handle scheme
    if keystore is None:
        taf_directory = find_taf_directory(Path())
        if taf_directory:
            keystore = str(taf_directory / "keystore")
        else:
            keystore = "./keystore"

    taf_logger.log("NOTICE", f"Generating keys in {str(Path(keystore).absolute())}")
    roles_key_infos_dict, keystore, _ = _initialize_roles_and_keystore(
        roles_key_infos, str(keystore)
    )

    roles_keys_data = from_dict(roles_key_infos_dict, RolesKeysData)
    for role in RolesIterator(roles_keys_data.roles, include_delegations=True):
        if not role.is_yubikey:
            for key_num in range(role.number):
                key_name = get_key_name(role.name, key_num, role.number)
                if keystore is not None:
                    password = input(
                        f"Enter {key_name} keystore password and press ENTER (can be left empty)"
                    )
                    key_path = str(Path(keystore, key_name))
                    _generate_rsa_key(key_path, password, role.length)
                else:
                    rsa_key, _ = generate_rsa_keypair(role.length)
                    print(f"{role.name} key:\n\n{rsa_key.decode()}\n\n")



"""TUF metadata repository"""


from fnmatch import fnmatch
from functools import reduce
import json
import operator
import os
from pathlib import Path
import logging
from collections import defaultdict
from datetime import datetime, timedelta, timezone
import shutil
from typing import Dict, List, Optional, Set, Tuple, Union
from securesystemslib.exceptions import StorageError
from cryptography.hazmat.primitives import serialization

from securesystemslib.signer import Signer
from securesystemslib import hash as sslib_hash

from taf import YubikeyMissingLibrary

from securesystemslib.storage import FilesystemBackend

from tuf.api.metadata import Signed

try:
    import taf.yubikey as yk
except ImportError:
    yk = YubikeyMissingLibrary()  # type: ignore

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.utils import (
    default_backend,
    get_file_details,
    on_rm_error,
    normalize_file_line_endings,
)
from tuf.api.metadata import (
    Metadata,
    MetaFile,
    Role,
    Root,
    Snapshot,
    Targets,
    TargetFile,
    Timestamp,
    DelegatedRole,
    Delegations,
)
from tuf.api.serialization.json import JSONSerializer
from taf.exceptions import InvalidKeyError, SignersNotLoaded, TAFError, TargetsError
from taf.models.types import RolesIterator, RolesKeysData, TargetsRole
from taf.tuf.keys import SSlibKey, _get_legacy_keyid, get_sslib_key_from_value
from tuf.repository import Repository

from securesystemslib.signer import CryptoSigner


logger = logging.getLogger(__name__)

# TODO remove this, use from constants or remove from constants
METADATA_DIRECTORY_NAME = "metadata"
TARGETS_DIRECTORY_NAME = "targets"

MAIN_ROLES = ["root", "targets", "snapshot", "timestamp"]

DISABLE_KEYS_CACHING = False
HASH_FUNCTION = "sha256"
HASH_ALGS = ["sha256", "sha512"]


def get_role_metadata_path(role: str) -> str:
    """
    Arguments:
        role: Name of a TUF role, main or delegated

    Return:
        Path of the metadata file corresponding to the specified role,
        relative to the repository's root
    """
    return f"{METADATA_DIRECTORY_NAME}/{role}.json"


def get_target_path(target_name: str) -> str:
    """
    Arguments:
        target_name: Name of the target file expected to be inside the targets directory

    Return:
        Path of the specified target file relative to the repository's root
    """
    return f"{TARGETS_DIRECTORY_NAME}/{target_name}"


def is_delegated_role(role: str) -> bool:
    return role not in ("root", "targets", "snapshot", "timestamp")


def is_auth_repo(repo_path: str) -> bool:
    """Check if the given path contains a valid TUF repository"""
    try:
        Repository(repo_path)._repository
        return True
    except Exception:
        return False


class MetadataRepository(Repository):
    """TUF metadata repository implementation for on-disk top-level roles.

    Provides methods to read and edit metadata, handling version and expiry
    bumps, and signature creation, and facilitating snapshot and timestamp
    creation.

    Arguments:
        path: Base path of metadata repository.

    Attributes:
        signer_cache: All signers available to the repository. Keys are role
            names, values are lists of signers. On `close` each signer for a
            role is used to sign the related role metadata.
        metadata_to_keep_open: A set containing metadata whose version numbers should not
        be increased when saving to disk. This makes it possible to combine multiple updates
    """

    expiration_intervals = {"root": 365, "targets": 90, "snapshot": 7, "timestamp": 1}

    serializer = JSONSerializer(compact=False)

    def __init__(self, path: Union[Path, str], *args, **kwargs) -> None:
        storage_backend = kwargs.pop("storage", None)
        super().__init__(*args, **kwargs)
        self.signer_cache: Dict[str, Dict[str, Signer]] = defaultdict(dict)
        self.path = Path(path)

        self._snapshot_info = MetaFile(1)
        self._targets_infos: Dict[str, MetaFile] = defaultdict(lambda: MetaFile(1))
        if storage_backend:
            self.storage_backend = storage_backend
        else:
            self.storage_backend = FilesystemBackend()
        self._metadata_to_keep_open: Set[str] = set()

    @property
    def metadata_path(self) -> Path:
        """
        Full path of the metadata directory.
        """
        return self.path / METADATA_DIRECTORY_NAME

    @property
    def targets_path(self):
        """
        Full path of the targets directory.
        """
        return self.path / TARGETS_DIRECTORY_NAME

    @property
    def targets_infos(self) -> Dict[str, MetaFile]:
        """
        Tracks targets and root metadata changes, needed in `do_snapshot`
        """
        return self._targets_infos

    @property
    def snapshot_info(self) -> MetaFile:
        """
        Tracks snapshot metadata changes, needed in `do_timestamp`
        """
        return self._snapshot_info

    def calculate_hashes(self, md: Metadata, algorithms: List[str]) -> Dict:
        """
        Calculate hashes of the specified signed metadata after serializing
        it using the previously initialized serializer.
        Hashes are computed for each specified algorithm.

        Arguments:
            md: Signed metadata
            algorithms: A list of hash algorithms (e.g., 'sha256', 'sha512').
        Return:
            A dcitionary mapping algorithms and calculated hashes
        """
        hashes = {}
        data = md.to_bytes(serializer=self.serializer)
        for algo in algorithms:
            digest_object = sslib_hash.digest(algo)
            digest_object.update(data)

            hashes[algo] = digest_object.hexdigest()
        return hashes

    def calculate_length(self, md: Metadata) -> int:
        """
        Calculate length of the specified signed metadata after serializing
        it using the previously initialized serializer.

        Arguments:
            md: Signed metadata
        Return:
            Langth of the signed metadata
        """
        data = md.to_bytes(serializer=self.serializer)
        return len(data)

    def add_signers_to_cache(self, roles_signers: Dict):
        for role, signers in roles_signers.items():
            if self._role_obj(role):
                self._load_role_signers(role, signers)

    def all_target_files(self) -> Set:
        """
        Return a set of relative paths of all files inside the targets
        directory
        """
        targets = []
        # Assume self.targets_path is a Path object, or convert it if necessary
        base_path = Path(self.targets_path)

        for filepath in base_path.rglob("*"):
            if filepath.is_file():
                # Get the relative path to the base directory and convert it to a POSIX path
                relative_path = filepath.relative_to(base_path).as_posix()
                targets.append(relative_path)

        return set(targets)

    def add_metadata_keys(self, roles_keys: Dict[str, List]) -> Tuple[Dict, Dict, Dict]:
        """Add signer public keys for role to root and update signer cache without updating snapshot and timestamp.

        Return:
            added_keys, already_added_keys, invalid_keys
        """
        already_added_keys = defaultdict(list)
        invalid_keys = defaultdict(list)
        added_keys = defaultdict(list)

        def _filter_if_can_be_added(roles):
            keys_to_be_added = defaultdict(list)
            for role, keys in roles_keys.items():
                if role in roles:
                    for key in keys:
                        try:
                            if self.is_valid_metadata_key(role, key):
                                already_added_keys[role].append(key)
                                continue
                        except TAFError:
                            invalid_keys[role].append(key)
                            continue
                        keys_to_be_added[role].append(key)
            return keys_to_be_added

        parents = self.find_parents_of_roles(list(roles_keys.keys()))
        self.verify_signers_loaded(parents)

        # when a key is added to one of the main roles
        # root is modified
        keys_to_be_added_to_root = _filter_if_can_be_added(MAIN_ROLES)
        if keys_to_be_added_to_root:
            with self.edit_root() as root:
                for role, keys in keys_to_be_added_to_root.items():
                    for key in keys:
                        root.add_key(key, role)
                        added_keys[role].append(key)

        other_roles = [role for role in roles_keys if role not in MAIN_ROLES]
        keys_to_be_added_to_targets = _filter_if_can_be_added(other_roles)

        roles_by_parents = defaultdict(list)
        if keys_to_be_added_to_targets:
            # group other roles by parents
            for role, keys in keys_to_be_added_to_targets.items():
                parent = self.find_delegated_roles_parent(role)
                roles_by_parents[parent].append(role)

            for parent, roles in roles_by_parents.items():
                with self.edit(parent) as parent_role:
                    for role in roles:
                        keys = roles_keys[role]
                        for key in keys:
                            parent_role.add_key(key, role)
                            added_keys[role].append(key)

        return added_keys, already_added_keys, invalid_keys

    def add_target_files_to_role(self, added_data: Dict[str, Dict]) -> None:
        """Add target files to top-level targets metadata.

        Arguments:
            added_data(dict): Dictionary of new data whose keys are target paths of repositories
            (as specified in targets.json, relative to the targets dictionary).
            The values are of form:
                {
                    target: content of the target file
                    custom: {
                        custom_field1: custom_value1,
                        custom_field2: custom_value2
                    }
                }
        """
        self.modify_targets(added_data=added_data)

    def add_path_to_delegated_role(self, role: str, paths: List[str]) -> bool:
        """
        Add delegated paths to delegated role and return True if successful

        Arguments:
            role: Name of a delegated target role
            path: A list of paths to be appended to a list of the role's delegated pats
        """
        if not self.check_if_role_exists(role):
            raise TAFError(f"Role {role} does not exist")

        parent_role = self.find_delegated_roles_parent(role)
        if parent_role is None:
            return False
        if all(
            path in self.get_delegations_of_role(parent_role)[role].paths
            for path in paths
        ):
            return False
        if parent_role:
            self.verify_signers_loaded([parent_role])
            with self.edit(parent_role) as parent:
                parent.delegations.roles[role].paths.extend(paths)
            return True
        return False

    def add_new_roles_to_snapshot(self, roles: List[str]) -> None:
        """
        Add versions of newly created target roles to the snapshot.
        Also update the versions of their parent roles, which are modified
        when a new delegated role is added.
        """
        with self.edit(Snapshot.type) as sn:
            parents_of_roles = set()
            for role in roles:
                sn.meta[f"{role}.json"] = MetaFile(1)
                parent_role = self.find_delegated_roles_parent(role)
                parents_of_roles.add(parent_role)
            for parent_role in parents_of_roles:
                sn.meta[f"{parent_role}.json"].version = (
                    sn.meta[f"{parent_role}.json"].version + 1
                )

    def add_to_open_metadata(self, roles: List[str]) -> None:
        """
        In order to execute several methods before updating the metadata on disk,
        some metadata might need to be kept open, which is done by adding them to
        _metadata_to_keep_open list.
        This method adds all roles from the provided list to _metadata_to_keep_open.
        """
        self._metadata_to_keep_open.update(roles)

    def open(self, role: str) -> Metadata:
        """Read role metadata from disk."""
        try:
            path = self.metadata_path / f"{role}.json"
            return Metadata.from_file(path, storage_backend=self.storage_backend)
        except StorageError:
            raise TAFError(f"Metadata file {path} does not exist")

    def check_if_keys_loaded(self, role_name: str) -> bool:
        """
        Check if at least a threshold of signers of the specified role
        has been added to the signer cache.
        """
        threshold = self.get_role_threshold(role_name)
        return (
            role_name in self.signer_cache
            and len(self.signer_cache[role_name]) >= threshold
        )

    def check_if_role_exists(self, role_name: str) -> bool:
        """
        Given a name of a main or delegated target role, return True if it exist
        """
        role = self._role_obj(role_name)
        return role is not None

    def check_roles_expiration_dates(
        self,
        interval: Optional[int] = None,
        start_date: Optional[datetime] = None,
        excluded_roles: Optional[List[str]] = None,
    ) -> Tuple[Dict, Dict]:
        """Determines which metadata roles have expired, or will expire within a time frame.
        Args:
        - interval(int): Number of days to look ahead for expiration.
        - start_date(datetime): Start date to look for expiration.
        - excluded_roles(list): List of roles to exclude from the search.

        Returns:
        - A dictionary of roles that have expired, or will expire within the given time frame.
        Results are sorted by expiration date.
        """
        if start_date is None:
            start_date = datetime.now(timezone.utc)
        if interval is None:
            interval = 30
        expiration_threshold = start_date + timedelta(days=interval)

        if excluded_roles is None:
            excluded_roles = []

        target_roles = self.get_all_targets_roles()
        main_roles = ["root", "targets", "snapshot", "timestamp"]
        existing_roles = list(set(target_roles + main_roles) - set(excluded_roles))

        expired_dict = {}
        will_expire_dict = {}
        for role in existing_roles:
            expiry_date = self.get_expiration_date(role)
            if start_date > expiry_date:
                expired_dict[role] = expiry_date
            elif expiration_threshold >= expiry_date:
                will_expire_dict[role] = expiry_date
        # sort by expiry date
        expired_dict = {
            k: v for k, v in sorted(expired_dict.items(), key=lambda item: item[1])
        }
        will_expire_dict = {
            k: v for k, v in sorted(will_expire_dict.items(), key=lambda item: item[1])
        }

        return expired_dict, will_expire_dict

    def _create_target_file(self, target_path: Path, target_data: Dict) -> None:
        """
        Writes the specified data to a target file and stores it on disk.
        Target data is of the following form:
        {
            target: content of the target file, string or Dict (json)
            custom: {
                custom_field1: custom_value1,
                custom_field2: custom_value2
            }
        }
        """
        # if the target's parent directory should not be "targets", create
        # its parent directories if they do not exist
        target_dir = target_path.parent
        target_dir.mkdir(parents=True, exist_ok=True)

        # create the target file
        content = target_data.get("target", None)
        if content is None:
            if not target_path.is_file():
                target_path.touch()
        else:
            with open(str(target_path), "w") as f:
                if isinstance(content, dict):
                    json.dump(content, f, indent=4)
                else:
                    f.write(content)

    def clear_open_metadata(self) -> None:
        """
        Removes everything from the _metadata_to_keep_open list
        """
        self._metadata_to_keep_open = set()

    def close(self, role: str, md: Metadata) -> None:
        """Bump version and expiry, re-sign, and write role metadata to disk."""

        # expiration date is updated before close is called
        if role not in self._metadata_to_keep_open:
            md.signed.version += 1

        md.signatures.clear()
        for signer in self.signer_cache[role].values():
            md.sign(signer, append=True)

        fname = f"{role}.json"

        # Track snapshot, targets and root metadata changes, needed in
        # `do_snapshot` and `do_timestamp`
        if role == "snapshot":
            self._snapshot_info.version = md.signed.version
            self._snapshot_info.hashes = self.calculate_hashes(md, HASH_ALGS)
            self._snapshot_info.length = self.calculate_length(md)
            root_version = self.signed_obj("root").version
            md.signed.meta["root.json"].version = root_version

        elif role != "timestamp":  # role in [root, targets, <delegated targets>]
            self._targets_infos[fname].version = md.signed.version

        # Write role metadata to disk (root gets a version-prefixed copy)
        md.to_file(
            self.metadata_path / fname,
            serializer=self.serializer,
            storage_backend=self.storage_backend,
        )

        if role == "root":
            md.to_file(
                self.metadata_path / f"{md.signed.version}.{fname}",
                serializer=self.serializer,
            )

    def create(
        self,
        roles_keys_data: RolesKeysData,
        signers: dict,
        additional_verification_keys: Optional[dict] = None,
    ) -> None:
        """Create a new metadata repository on disk.

        1. Create metadata subdir (fail, if exists)
        2. Create initial versions of top-level metadata
        3. Perform top-level delegation using keys from passed signers.

        Arguments:
            roles_keys_data: an object containing information about roles, their threshold, delegations etc.
            signers: A dictionary, where dict-keys are role names and values
                are dictionaries, where-dict keys are keyids and values
                are signers.
            additional_verification_keys: A dictionary where keys are names of roles and values are lists
                of public keys that should be registered as the corresponding role's keys, but the private
                keys are not available. E.g. keys exporeted from YubiKeys of maintainers who are not
                present at the time of the repository's creation
        """
        # TODO add verification keys
        # support yubikeys
        self.metadata_path.mkdir(parents=True)
        self.signer_cache = defaultdict(dict)

        root = Root(consistent_snapshot=False)

        # Snapshot tracks targets and root versions. targets v1 is included by
        # default in snapshot v1. root must be added explicitly.
        sn = Snapshot()
        sn.meta["root.json"] = MetaFile(1)

        public_keys = {
            role_name: {
                _get_legacy_keyid(signer.public_key): signer.public_key
                for signer in role_signers
            }
            for role_name, role_signers in signers.items()
        }
        if additional_verification_keys:
            for role_name, roles_public_keys in additional_verification_keys.items():
                for public_key in roles_public_keys:
                    key_id = _get_legacy_keyid(public_key)
                    if key_id not in public_keys[role_name]:
                        public_keys[role_name][key_id] = public_key

        for role in RolesIterator(roles_keys_data.roles, include_delegations=False):
            if signers.get(role.name) is None:
                raise TAFError(f"Cannot setup role {role.name}. Keys not specified")
            for signer in signers[role.name]:
                key_id = _get_legacy_keyid(signer.public_key)
                self.signer_cache[role.name][key_id] = signer
            for public_key in public_keys[role.name].values():
                root.add_key(public_key, role.name)
            root.roles[role.name].threshold = role.threshold

        targets = Targets()
        target_roles = {"targets": targets}
        delegations_per_parent: Dict[str, Dict] = defaultdict(dict)
        for role in RolesIterator(roles_keys_data.roles.targets):
            if role.parent is None:
                continue
            parent = role.parent.name
            parent_obj = target_roles.get(parent)
            for signer in signers[role.name]:
                self.signer_cache[role.name][key_id] = signer
            delegated_role = DelegatedRole(
                name=role.name,
                threshold=role.threshold,
                paths=role.paths,
                terminating=role.terminating,
                keyids=list(public_keys[role.name].keys()),
            )
            delegated_metadata = Targets()
            target_roles[role.name] = delegated_metadata
            delegations_per_parent[parent][role.name] = delegated_role
            sn.meta[f"{role.name}.json"] = MetaFile(1)

        for parent, role_data in delegations_per_parent.items():
            parent_obj = target_roles[parent]
            delegated_keys = {}
            for delegated_role_name in role_data:
                delegated_keys.update(public_keys[delegated_role_name])
            delegations = Delegations(roles=role_data, keys=delegated_keys)
            parent_obj.delegations = delegations

        for signed in [root, Timestamp(), sn, targets]:
            # Setting the version to 0 here is a trick, so that `close` can
            # always bump by the version 1, even for the first time
            self._set_default_expiration_date(signed)
            signed.version = 0  # `close` will bump to initial valid verison 1
            self.close(signed.type, Metadata(signed))

        for name, signed in target_roles.items():
            if name != "targets":
                self._set_default_expiration_date(signed)
                signed.version = 0  # `close` will bump to initial valid verison 1
                self.close(name, Metadata(signed))

    def create_delegated_roles(
        self, roles_data: List[TargetsRole], signers: Dict[str, List[CryptoSigner]]
    ) -> Tuple[List, List]:
        """
        Create a new delegated roles, signes them using the provided signers and
        updates their paren roles.

        Arguments:
            roles_data (list): A list containing data about new roles. Each entry specifies
                            a role's name, path, threshold, and number of signing keys.
            signers (dict): A dictionary that maps each new role to a list of its signers

        Return:
            A list ofroles that were added and a list of roles that already existed
        """
        existing_roles = self.get_all_targets_roles()
        existing_roles.extend(MAIN_ROLES)
        existing_roles = []
        added_roles = []
        roles_parents_dict = defaultdict(list)
        for role_data in roles_data:
            if role_data.name in existing_roles:
                existing_roles.append(role_data.name)
                continue
            if role_data.parent is not None:
                parent = role_data.parent.name
                roles_parents_dict[parent].append(role_data)

        for parent, parents_roles_data in roles_parents_dict.items():
            with self.edit(parent) as parent_obj:
                keys_data = {}
                for role_data in parents_roles_data:
                    for signer in signers[role_data.name]:
                        public_key = signer.public_key
                        key_id = _get_legacy_keyid(public_key)
                        keys_data[key_id] = public_key
                        self.signer_cache[role_data.name][key_id] = signer
                    delegated_role = DelegatedRole(
                        name=role_data.name,
                        threshold=role_data.threshold,
                        paths=role_data.paths,
                        terminating=role_data.terminating,
                        keyids=list(keys_data.keys()),
                    )
                    if parent_obj.delegations is None:
                        parent_obj.delegations = Delegations(
                            roles={role_data.name: delegated_role}, keys=keys_data
                        )
                    else:
                        parent_obj.delegations.roles[role_data.name] = delegated_role
                parent_obj.delegations.keys.update(keys_data)

            for role_data in parents_roles_data:
                new_role_signed = Targets()
                self._set_default_expiration_date(new_role_signed)
                new_role_signed.version = (
                    0  # `close` will bump to initial valid verison 1
                )
                self.close(role_data.name, Metadata(new_role_signed))
                added_roles.append(role_data.name)
        return added_roles, existing_roles

    def create_and_remove_target_files(
        self, added_data: Optional[Dict] = None, removed_data: Optional[Dict] = None
    ) -> Tuple:
        """Create/updates/removes files in the targets directory
        Args:
        - added_data(dict): Dictionary of new data whose keys are target paths of repositories
                            (as specified in targets.json, relative to the targets dictionary).
                            The values are of form:
                            {
                                target: content of the target file
                            }
        - removed_data(dict): Dictionary of the old data whose keys are target paths of
                              repositories
                              (as specified in targets.json, relative to the targets dictionary).
                              The values are not needed. This is just for consistency.

        Content of the target file can be a dictionary, in which case a json file will be created.
        If that is not the case, an ordinary textual file will be created.
        If content is not specified and the file already exists, it will not be modified.
        If it does not exist, an empty file will be created. To replace an existing file with an
        empty file, specify empty content (target: '')

        Returns:
        - Role whose targets were updates
        """
        added_data = {} if added_data is None else added_data
        removed_data = {} if removed_data is None else removed_data
        data = dict(added_data, **removed_data)
        if not data:
            raise TargetsError("Nothing to be modified!")

        added_paths = []
        for path, target_data in added_data.items():
            target_path = (self.targets_path / path).absolute()
            self._create_target_file(target_path, target_data)
            added_paths.append(target_path)

        # remove existing target files
        removed_paths = []
        for path in removed_data.keys():
            target_path = (self.targets_path / path).absolute()
            if target_path.exists():
                if target_path.is_file():
                    target_path.unlink()
                elif target_path.is_dir():
                    shutil.rmtree(target_path, onerror=on_rm_error)
            removed_paths.append(str(path))

        return added_paths, removed_paths

    def _create_target_object(
        self, filesystem_path: str, target_path: str, custom: Optional[Dict]
    ) -> TargetFile:
        """
        Creates a TUF target object, later used to update targets metadata.
        It's first necessary to normalize file line endings (convert all line endings to unix style endings)
        before adding a target objects due to hashes getting calculated differently when using CRLF vs LF line endings.
        So we instead convert all to unix style endings.
        """
        normalize_file_line_endings(filesystem_path)
        data = Path(filesystem_path).read_text().encode()
        target_file = TargetFile.from_data(
            target_file_path=target_path,
            data=data,
            hash_algorithms=["sha256", "sha512"],
        )
        if custom:
            unrecognized_fields = {"custom": custom}
            target_file.unrecognized_fields = unrecognized_fields
        return target_file

    def delete_unregistered_target_files(self, targets_role="targets"):
        """
        Delete all target files not specified in targets.json
        """
        target_files_by_roles = self.sort_roles_targets_for_filenames()
        if targets_role in target_files_by_roles:
            for file_rel_path in target_files_by_roles[targets_role]:
                if file_rel_path not in self.get_targets_of_role(targets_role):
                    (self.targets_path / file_rel_path).unlink()

    def do_timestamp(self, force=False):
        self._snapshot_info.version = self._signed_obj(Snapshot.type).version
        return super().do_timestamp(force)

    def find_delegated_roles_parent(self, delegated_role: str) -> Optional[str]:
        """
        Find parent role of the specified delegated targets role
        """
        parents = ["targets"]

        while parents:
            parent = parents.pop()
            for delegation in self.get_delegations_of_role(parent):
                if delegation == delegated_role:
                    return parent
                parents.append(delegation)
        return None

    def find_parents_of_roles(self, roles: List[str]):
        """
        Find parents of all roles contained by the specified list of roles.
        """
        parents = set()
        for role in roles:
            if role in MAIN_ROLES:
                parents.add("root")
            else:
                parent = self.find_delegated_roles_parent(role)
                if parent is None:
                    raise TAFError(f"Could not determine parent of role {role}")
                parents.add(parent)
        return parents

    def get_delegations_of_role(self, role_name: str) -> Dict:
        """
        Return a dictionary of delegated roles of the specified target role
        """
        signed_obj = self.signed_obj(role_name)
        if signed_obj.delegations:
            return signed_obj.delegations.roles
        return {}

    def get_keyids_of_role(self, role_name: str) -> List:
        """
        Return all key ids of the specified role
        """
        role_obj = self._role_obj(role_name)
        return role_obj.keyids

    def get_paths_of_role(self, role_name: str) -> List:
        """
        Return all delegated paths of the specified target role
        """
        parent = self.find_delegated_roles_parent(role_name)
        if parent:
            parent_obj = self.signed_obj(parent)
            return parent_obj.delegations.roles[role_name].paths
        return []

    def get_targets_of_role(self, role_name: str):
        """
        Return all targets of the specified target role
        """
        return self.signed_obj(role_name).targets

    def find_keys_roles(
        self, public_keys: List, check_threshold: Optional[bool] = True
    ) -> List:
        """Find all roles that can be signed by the provided keys.
        A role can be signed by the list of keys if at least the number
        of keys that can sign that file is equal to or greater than the role's
        threshold
        """
        key_ids = [_get_legacy_keyid(public_key) for public_key in public_keys]
        return self.find_keysid_roles(key_ids=key_ids, check_threshold=check_threshold)

    def find_keysid_roles(
        self, key_ids: List, check_threshold: Optional[bool] = True
    ) -> List:
        """Find all roles that can be signed by the provided keys.
        A role can be signed by the list of keys if at least the number
        of keys that can sign that file is equal to or greater than the role's
        threshold
        """
        roles: List[Tuple[str, Optional[str]]] = []
        for role in MAIN_ROLES:
            roles.append((role, None))
        keys_roles = []
        while roles:
            role_name, parent = roles.pop()
            role_obj = self._role_obj(role_name, parent)
            target_roles_key_ids = role_obj.keyids
            threshold = role_obj.threshold
            num_of_signing_keys = len(set(target_roles_key_ids).intersection(key_ids))
            if (
                not check_threshold and num_of_signing_keys >= 1
            ) or num_of_signing_keys >= threshold:
                keys_roles.append(role_name)

            if role_name not in MAIN_ROLES or role_name == "targets":
                for delegation in self.get_delegations_of_role(role_name):
                    roles.append((delegation, role_name))

        return keys_roles

    def find_associated_roles_of_key(self, public_key: SSlibKey) -> List:
        """
        Find all roles whose metadata files can be signed by this key
        Threshold is not important, as long as the key is one of the signing keys
        """
        return self.find_keys_roles([public_key], check_threshold=False)

    def get_all_roles(self) -> List:
        """
        Return a list of all defined roles, main roles combined with delegated targets roles
        """
        all_target_roles = self.get_all_targets_roles()
        all_roles = ["root", "snapshot", "timestamp"] + all_target_roles
        return all_roles

    def get_all_targets_roles(self) -> List:
        """
        Return a list containing names of all target roles
        """
        target_roles = ["targets"]
        all_roles = []
        while target_roles:
            role = target_roles.pop()
            all_roles.append(role)
            for delegation in self.get_delegations_of_role(role):
                target_roles.append(delegation)

        return all_roles

    def get_all_target_files_state(self) -> Tuple:
        """Create dictionaries of added/modified and removed files by comparing current
        file-system state with current signed targets (and delegations) metadata state.

        Args:
        - None
        Returns:
        - Dict of added/modified files and dict of removed target files (inputs for
          `modify_targets` method.)

        Raises:
        - None
        """
        added_target_files: Dict = {}
        removed_target_files: Dict = {}
        # current fs state
        fs_target_files = self.all_target_files()
        # current signed state
        signed_target_files = self.get_signed_target_files()

        # existing files with custom data and (modified) content
        for file_name in fs_target_files:
            target_file = self.targets_path / file_name
            _, hashes = get_file_details(str(target_file))
            # register only new or changed files
            if hashes.get(HASH_FUNCTION) != self.get_target_file_hashes(file_name):
                custom = self.get_target_file_custom_data(file_name)
                added_target_files[file_name] = {
                    "target": target_file.read_text(),
                }
                if custom:
                    added_target_files[file_name]["custom"] = custom

        # removed files
        for file_name in signed_target_files - fs_target_files:
            removed_target_files[file_name] = {}

        return added_target_files, removed_target_files

    def get_expiration_date(self, role: str) -> datetime:
        """
        Return expiration date of the specified role
        """
        meta_file = self.signed_obj(role)
        if meta_file is None:
            raise TAFError(f"Role {role} does not exist")

        date = meta_file.expires
        return date.replace(tzinfo=timezone.utc)

    def get_role_threshold(self, role: str, parent: Optional[str] = None) -> int:
        """Get threshold of the given role

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - parent_role(str): Name of the parent role of the delegated role. If not specified,
                            it will be set automatically, but this might be slow if there
                            are many delegations.

        Returns:
        Role's signatures threshold

        Raises:
        - TAFError if the role does not exist or if metadata files are invalid
        """
        role_obj = self._role_obj(role, parent)
        if role_obj is None:
            raise TAFError(f"Role {role} does not exist")
        return role_obj.threshold

    def get_role_paths(self, role):
        """Get paths of the given role

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)

        Returns:
        Defined delegated paths of delegate target role or * in case of targets
        (the main target is responsible for signing all target files that no delegated role should sign.)

        Raises:
        - TAFError if the role does not exist
        """
        if role == "targets":
            return "*"
        role = self._role_obj(role)
        if role is None:
            raise TAFError(f"Role {role} does not exist")
        return role.paths

    def get_role_from_target_paths(self, target_paths: List) -> Optional[str]:
        """
        Find a common role that can be used to sign given target paths.

        NOTE: Currently each target has only one mapped role.
        """
        targets_roles = self.map_signing_roles(target_paths)
        roles = list(targets_roles.values())

        try:
            # all target files should have at least one common role
            common_role = reduce(
                set.intersection,
                [set([r]) if isinstance(r, str) else set(r) for r in roles],
            )
        except TypeError:
            return None

        if not common_role:
            return None

        return common_role.pop()

    def get_signable_metadata(self, role: str):
        """Return signable portion of newly generate metadata for given role.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)

        Returns:
        A string representing the 'object' encoded in canonical JSON form or None

        Raises:
        None
        """
        signed = self.signed_obj(role)
        return signed.to_dict()

    def get_signed_target_files(self) -> Set[str]:
        """Return all target files signed by all roles.

        Args:
        - None

        Returns:
        - Set of all target paths relative to targets directory
        """
        all_roles = self.get_all_targets_roles()
        return self.get_signed_target_files_of_roles(all_roles)

    def get_signed_target_files_of_roles(
        self, roles: Optional[List] = None
    ) -> Set[str]:
        """Return all target files signed by the specified roles

        Args:
        - roles whose target files will be returned

        Returns:
        - Set of paths of target files of a role relative to targets directory
        """
        if roles is None:
            roles = self.get_all_targets_roles()

        return set(
            reduce(
                operator.iconcat,
                [self.signed_obj(role).targets.keys() for role in roles],
                [],
            )
        )

    def get_signed_targets_with_custom_data(
        self, roles: Optional[List[str]] = None
    ) -> Dict[str, Dict]:
        """Return all target files signed by the specified roles and and their custom data
        as specified in the metadata files

        Args:
        - roles whose target files will be returned

        Returns:
        - A dictionary whose keys are paths target files relative to the targets directory
        and values are custom data dictionaries.
        """
        if roles is None:
            roles = self.get_all_targets_roles()
        target_files: Dict[str, Dict] = {}
        try:
            for role in roles:
                roles_targets = self.get_targets_of_role(role)
                for target_path, target_file in roles_targets.items():
                    target_files.setdefault(target_path, {}).update(
                        target_file.custom or {}
                    )
        except StorageError:
            pass
        return target_files

    def get_target_file_custom_data(self, target_path: str) -> Optional[Dict]:
        """
        Return a custom data of a given target.
        """
        try:
            role = self.get_role_from_target_paths([target_path])
            if role is None:
                return None
            target_obj = self.get_targets_of_role(role).get(target_path)
            if target_obj:
                return target_obj.custom
            return None
        except KeyError:
            raise TAFError(f"Target {target_path} does not exist")

    def get_target_file_hashes(
        self, target_path: str, hash_func: str = HASH_FUNCTION
    ) -> Optional[str]:
        """
        Return hashes of the given target path.

        Raises:
        - TAFError if the target does not exist
        """
        try:
            role = self.get_role_from_target_paths([target_path])
            if role is None:
                return None
            targets_of_role = self.get_targets_of_role(role)
            if target_path not in targets_of_role:
                return None
            hashes = targets_of_role[target_path].hashes
            if hash_func not in hashes:
                raise TAFError(f"Invalid hashing algorithm {hash_func}")
            return hashes[hash_func]
        except KeyError:
            raise TAFError(f"Target {target_path} does not exist")

    def get_key_length_and_scheme_from_metadata(
        self, parent_role: str, keyid: str
    ) -> Tuple:
        """
        Return length and signing scheme of the specified key id.
        This data is specified in metadata files (root or a target role that has delegations)
        """
        try:
            metadata = json.loads(
                Path(
                    self.path, METADATA_DIRECTORY_NAME, f"{parent_role}.json"
                ).read_text()
            )
            metadata = metadata["signed"]
            if "delegations" in metadata:
                metadata = metadata["delegations"]
            scheme = metadata["keys"][keyid]["scheme"]
            pub_key_pem = metadata["keys"][keyid]["keyval"]["public"]
            pub_key = serialization.load_pem_public_key(
                pub_key_pem.encode(), backend=default_backend()
            )
            return pub_key, scheme
        except Exception:
            return None, None

    def generate_roles_description(self) -> Dict:
        """
        Generate a roles description dictionary, containing information
        about each role, like its threhold, number of signing keys, delegations
        if it is a target role, key scheme, key lengths.
        """
        roles_description = {}

        def _get_delegations(role_name):
            delegations_info = {}
            for delegation in self.get_delegations_of_role(role_name):
                delegated_role = self._role_obj(delegation)
                delegations_info[delegation] = {
                    "threshold": delegated_role.threshold,
                    "number": len(delegated_role.keyids),
                    "paths": delegated_role.paths,
                    "terminating": delegated_role.terminating,
                }
                pub_key, scheme = self.get_key_length_and_scheme_from_metadata(
                    role_name, delegated_role.keyids[0]
                )

                delegations_info[delegation]["scheme"] = scheme
                delegations_info[delegation]["length"] = pub_key.key_size
                delegated_signed = self.signed_obj(delegation)
                if delegated_signed.delegations:
                    inner_roles_data = _get_delegations(delegation)
                    if len(inner_roles_data):
                        delegations_info[delegation]["delegations"] = inner_roles_data
            return delegations_info

        for role_name in MAIN_ROLES:
            role_obj = self._role_obj(role_name)
            roles_description[role_name] = {
                "threshold": role_obj.threshold,
                "number": len(role_obj.keyids),
            }
            pub_key, scheme = self.get_key_length_and_scheme_from_metadata(
                "root", role_obj.keyids[0]
            )
            roles_description[role_name]["scheme"] = scheme
            roles_description[role_name]["length"] = pub_key.key_size
            if role_name == "targets":
                targets_signed = self.signed_obj(role_name)
                if targets_signed.delegations:
                    delegations_info = _get_delegations(role_name)
                    if len(delegations_info):
                        roles_description[role_name]["delegations"] = delegations_info
        return {"roles": roles_description}

    def get_role_keys(self, role):
        """Get keyids of the given role

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)

        Returns:
            List of the role's keyids (i.e., keyids of the keys).

        """
        role_obj = self._role_obj(role)
        if role_obj is None:
            return None
        try:
            return role_obj.keyids
        except KeyError:
            pass

    def is_valid_metadata_key(
        self, role: str, key: Union[SSlibKey, str], scheme=DEFAULT_RSA_SIGNATURE_SCHEME
    ) -> bool:
        """Checks if metadata role contains key id of provided key.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - key(securesystemslib.formats.RSAKEY_SCHEMA): Role's key.

        Returns:
        Boolean. True if key id is in metadata role key ids, False otherwise.

        Raises:
        - TAFError if key is not valid
        """

        try:
            if isinstance(key, str):
                # mypy will complain if we redefine key
                ssl_lib_key = get_sslib_key_from_value(key, scheme)
            else:
                ssl_lib_key = key
            key_id = _get_legacy_keyid(ssl_lib_key)
        except Exception:
            # TODO log
            raise TAFError("Invalid public key specified")
        else:
            return key_id in self.get_keyids_of_role(role)

    def is_valid_metadata_yubikey(self, role: str, public_key=None) -> bool:
        """Checks if metadata role contains key id from YubiKey.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one
        - public_key(securesystemslib.formats.RSAKEY_SCHEMA): RSA public key dict

        Returns:
        Boolean. True if smart card key id belongs to metadata role key ids

        Raises:
        - YubikeyError
        """

        if public_key is None:
            public_key = yk.get_piv_public_key_tuf()

        return self.is_valid_metadata_key(role, public_key)

    def _load_role_signers(self, role: str, signers: List) -> None:
        """Verify that the signers can be used to sign the specified role and
        add them to the signer cache

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - signers: A list of signers

        Returns:
        None

        Raises:
        - InvalidKeyError: If metadata cannot be signed with given key.
        """

        for signer in signers:
            key = signer.public_key
            if not self.is_valid_metadata_key(role, key):
                raise InvalidKeyError(role)
            self.signer_cache[role][key.keyid] = signer

    def map_signing_roles(self, target_filenames: List) -> Dict:
        """
        For each target file, find delegated role responsible for that target file based
        on the delegated paths. The most specific role (meaning most deeply nested) whose
        delegation path matches the target's path is returned as that file's matching role.
        If there are no delegated roles with a path that matches the target file's path,
        'targets' role will be returned as that file's matching role. Delegation path
        is expected to be relative to the targets directory. It can be defined as a glob
        pattern.
        """

        roles = ["targets"]
        roles_targets = {
            target_filename: "targets" for target_filename in target_filenames
        }
        while roles:
            role = roles.pop()
            path_patterns = self.get_role_paths(role)
            for path_pattern in path_patterns:
                for target_filename in target_filenames:
                    if fnmatch(
                        target_filename.lstrip(os.sep),
                        path_pattern.lstrip(os.sep),
                    ):
                        roles_targets[target_filename] = role

            for delegation in self.get_delegations_of_role(role):
                roles.append(delegation)

        return roles_targets

    def modify_targets(
        self, added_data: Optional[Dict] = None, removed_data: Optional[Dict] = None
    ) -> Targets:
        """Creates a target.json file containing a repository's commit for each repository.
        Adds those files to the tuf repository.

        Args:
        - added_data(dict): Dictionary of new data whose keys are target paths of repositories
                            (as specified in targets.json, relative to the targets dictionary).
                            The values are of form:
                            {
                                target: content of the target file
                                custom: {
                                    custom_field1: custom_value1,
                                    custom_field2: custom_value2
                                }
                            }
        - removed_data(dict): Dictionary of the old data whose keys are target paths of
                              repositories
                              (as specified in targets.json, relative to the targets dictionary).
                              The values are not needed. This is just for consistency.

        Content of the target file can be a dictionary, in which case a json file will be created.
        If that is not the case, an ordinary textual file will be created.
        If content is not specified and the file already exists, it will not be modified.
        If it does not exist, an empty file will be created. To replace an existing file with an
        empty file, specify empty content (target: '')

        Custom is an optional property which, if present, will be used to specify a TUF target's

        Returns:
        - Role whose targets were updates
        """
        added_data = {} if added_data is None else added_data
        removed_data = {} if removed_data is None else removed_data
        data = dict(added_data, **removed_data)
        if not data:
            raise TargetsError("Nothing to be modified!")

        target_paths = list(data.keys())
        targets_role = self.get_role_from_target_paths(target_paths)
        if targets_role is None:
            raise TargetsError(
                f"Could not find a common role for target paths:\n{'-'.join(target_paths)}"
            )
        _, removed_paths = self.create_and_remove_target_files(added_data, removed_data)

        target_files = []
        for path, target_data in added_data.items():
            target_path = (self.targets_path / path).absolute()
            custom = target_data.get("custom", None)
            target_file = self._create_target_object(target_path, path, custom)
            target_files.append(target_file)

        targets_role = self._modify_targets_role(
            target_files, removed_paths, targets_role
        )
        return targets_role

    def _modify_targets_role(
        self,
        added_target_files: List[TargetFile],
        removed_paths: List[str],
        role_name: Optional[str] = Targets.type,
    ) -> Targets:
        """Add target files to top-level targets metadata."""
        with self.edit_targets(rolename=role_name) as targets:
            for target_file in added_target_files:
                targets.targets[target_file.path] = target_file
            for path in removed_paths:
                targets.targets.pop(path, None)
        return targets

    def revoke_metadata_key(self, key_id: str, roles: Optional[List[str]] = None):
        """Remove metadata key of the provided role without updating timestamp and snapshot.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - key_id(str): An object conformant to 'securesystemslib.formats.KEYID_SCHEMA'.

        Returns:
            removed_from_roles, not_added_roles, less_than_threshold_roles
        """
        if key_id is None:
            raise TAFError("Keyid to revoke not specified")
        if not roles:
            roles = self.find_keysid_roles([key_id])

        if not roles:
            raise TAFError("Key not used to sign any role")

        parents = self.find_parents_of_roles(roles)
        self.verify_signers_loaded(parents)

        removed_from_roles = []
        not_added_roles = []
        less_than_threshold_roles = []

        def _check_if_can_remove(key_id, role):
            role_obj = self._role_obj(role)
            if len(role_obj.keyids) - 1 < role_obj.threshold:
                less_than_threshold_roles.append(role)
                return False
            key_ids_of_role = self.get_keyids_of_role(role) or []
            if key_id not in key_ids_of_role:
                not_added_roles.append(role)
                return False
            return True

        main_roles = [
            role
            for role in roles
            if role in MAIN_ROLES and _check_if_can_remove(key_id, role)
        ]
        if len(main_roles):
            with self.edit_root() as root:
                for role in main_roles:
                    root.revoke_key(keyid=key_id, role=role)
                    removed_from_roles.append(role)

        roles_by_parents = defaultdict(list)
        delegated_roles = [
            role
            for role in roles
            if role not in MAIN_ROLES and _check_if_can_remove(key_id, role)
        ]
        if len(delegated_roles):
            for role in delegated_roles:
                parent = self.find_delegated_roles_parent(role)
                roles_by_parents[parent].append(role)

            for parent, roles_of_parent in roles_by_parents.items():
                with self.edit(parent) as parent_role:
                    for role in roles_of_parent:
                        parent_role.revoke_key(keyid=key_id, role=role)
                        removed_from_roles.append(role)

        return removed_from_roles, not_added_roles, less_than_threshold_roles

    def remove_delegated_paths(self, roles_paths: Dict[str, List[str]]):
        """
        Remove delegated paths to delegated role and return True if at least one removed
        """

        updated = False
        for role, paths in roles_paths.items():

            if not self.check_if_role_exists(role):
                raise TAFError(f"Role {role} does not exist")

            parent_role = self.find_delegated_roles_parent(role)
            if parent_role is None:
                raise TAFError(f"Role {role} is not a delegated role")

            self.verify_signers_loaded([parent_role])
            with self.edit(parent_role) as parent:
                for path in paths:
                    if path in parent.delegations.roles[role].paths:
                        parent.delegations.roles[role].paths.remove(path)
                        updated = True
        return updated

    def remove_from_open_metadata(self, roles: List[str]) -> None:
        """
        Removes the listed roles from metadata_to_keep_open list
        """
        for role in roles:
            if role in self._metadata_to_keep_open:
                self._metadata_to_keep_open.remove(role)

    def roles_targets_for_filenames(self, target_filenames):
        """Sort target files by roles
        Args:
        - target_filenames: List of relative paths of target files
        Returns:
        - A dictionary mapping roles to a list of target files belonging
          to the provided target_filenames list delegated to the role
        """
        targets_roles_mapping = self.map_signing_roles(target_filenames)
        roles_targets_mapping = {}
        for target_filename, role_name in targets_roles_mapping.items():
            roles_targets_mapping.setdefault(role_name, []).append(target_filename)
        return roles_targets_mapping

    def _role_obj(self, role: str, parent: Optional[str] = None):
        """
        Return TUF's role object for the specified role
        """
        if role in MAIN_ROLES:
            md = self.open("root")
            try:
                data = md.to_dict()["signed"]["roles"][role]
                return Role.from_dict(data)
            except (KeyError, ValueError):
                raise TAFError("root.json is invalid")
        else:
            parent_name = self.find_delegated_roles_parent(role)
            if parent_name is None:
                return None
            md = self.open(parent_name)
            delegations_data = md.to_dict()["signed"]["delegations"]["roles"]
            for delegation in delegations_data:
                if delegation["name"] == role:
                    try:
                        return DelegatedRole.from_dict(delegation)
                    except (KeyError, ValueError):
                        raise TAFError(f"{delegation}.json is invalid")
            return None

    def signed_obj(self, role: str):
        """
        Return TUF's signed object for the specified role
        """
        md = self.open(role)
        return self._signed_obj(role, md)

    def _signed_obj(self, role: str, md=None):
        if md is None:
            md = self.open(role)
        try:
            signed_data = md.to_dict()["signed"]
            role_to_role_class = {
                "root": Root,
                "targets": Targets,
                "snapshot": Snapshot,
                "timestamp": Timestamp,
            }
            role_class = role_to_role_class.get(role, Targets)
            return role_class.from_dict(signed_data)
        except (KeyError, ValueError):
            raise TAFError(f"Invalid metadata file {role}.json")

    def _set_default_expiration_date(self, signed: Signed) -> None:
        """
        Update expiration dates of the specified signed object
        """
        interval = self.expiration_intervals.get(signed.type, 90)
        start_date = datetime.now(timezone.utc)
        expiration_date = start_date + timedelta(days=interval)
        signed.expires = expiration_date

    def set_metadata_expiration_date(
        self,
        role_name: str,
        start_date: Optional[datetime] = None,
        interval: Optional[int] = None,
    ) -> None:
        """Set expiration date of the provided role.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - start_date(datetime): Date to which the specified interval is added when calculating
                                expiration date. If a value is not provided, it is set to the
                                current time.
        - signers(List[CryptoSigner]): a list of signers
        - interval(int): A number of days added to the start date.
                        If not provided, the default value is set based on the role:

                            root - 365 days
                            targets - 90 days
                            snapshot - 7 days
                            timestamp - 1 day
                            all other roles (delegations) - same as targets

        Returns:
        None

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by
                                                        this targets object.
        """
        self.verify_signers_loaded([role_name])
        with self.edit(role_name) as role:
            start_date = datetime.now(timezone.utc)
            if interval is None:
                try:
                    interval = self.expiration_intervals[role_name]
                except KeyError:
                    interval = self.expiration_intervals["targets"]
            expiration_date = start_date + timedelta(days=interval)
            role.expires = expiration_date

    def sort_roles_targets_for_filenames(self):
        """
        Group target files per target roles
        """
        rel_paths = []
        for filepath in self.targets_path.rglob("*"):
            if filepath.is_file():
                file_rel_path = str(
                    Path(filepath).relative_to(self.targets_path).as_posix()
                )
                rel_paths.append(file_rel_path)

        files_to_roles = self.map_signing_roles(rel_paths)
        roles_targets = {}
        for target_file, role in files_to_roles.items():
            roles_targets.setdefault(role, []).append(target_file)
        return roles_targets

    def sync_snapshot_with_roles(self, roles: List[str]) -> None:
        """
        Add versions of newly created target roles to the snapshot.
        Also update the versions of their parent roles, which are modified
        when a new delegated role is added.
        """
        with self.edit(Snapshot.type) as sn:
            for role in roles:
                sn.meta[f"{role}.json"].version = sn.meta[f"{role}.json"].version + 1

    def update_target_role(self, role: str, target_paths: Dict, force=False):
        """
        Update the specified target role by adding or removing
        target files and target objects for the specified target paths
        If false is True, update the metadata files even if no target
        paths are specified
        """
        if not self.check_if_role_exists(role):
            raise TAFError(f"Role {role} does not exist")
        self.verify_signers_loaded([role])
        removed_paths = []
        target_files = []
        if target_paths:
            for target_path in target_paths:
                full_path = self.path / TARGETS_DIRECTORY_NAME / target_path
                # file removed, removed from te role
                if not full_path.is_file():
                    removed_paths.append(target_path)
                else:
                    custom_data = self.get_target_file_custom_data(target_path)
                    target_file = self._create_target_object(
                        full_path, target_path, custom_data
                    )
                    target_files.append(target_file)

            self._modify_targets_role(target_files, removed_paths, role)
        elif force:
            with self.edit(role) as _:
                pass

    def update_snapshot_and_timestamp(self, force: Optional[bool] = True):
        """
        Update timestamp and snapshot roles. If force is true, update them
        even if their content was not modified
        """
        self.verify_signers_loaded(["snapshot", "timestamp"])
        self.do_snapshot(force=force)
        self.do_timestamp(force=force)

    def verify_roles_exist(self, roles: List[str]):
        """
        Check if the specified roles exist and raise an error if an least one does not exist
        """
        non_existant_roles = []
        for role in roles:
            if not self.check_if_role_exists(role):
                non_existant_roles.append(role)
        if len(non_existant_roles):
            raise TAFError(f"Role(s) {', '.join(non_existant_roles)} do not exist")

    def verify_signers_loaded(self, roles: List[str]):
        """
        Verify that the signers associated with the specified keys were added to the signer cache.
        Raise an error if that is not the case
        """
        not_loaded = [role for role in roles if role not in self.signer_cache]
        if len(not_loaded):
            raise SignersNotLoaded(roles=not_loaded)



import json
from logging import DEBUG, ERROR
from typing import Dict, Optional
import click
from taf.api.targets import register_target_files
import taf.repositoriesdb as repositoriesdb
from logdecorator import log_on_end, log_on_error, log_on_start
from taf.api.utils._git import check_if_clean
from taf.messages import git_commit_message
from pathlib import Path

from taf.auth_repo import AuthenticationRepository
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import TAFError
from taf.git import GitRepository
from taf.log import taf_logger
from taf.updater.updater import OperationType, clone_repository
import taf.updater.updater as updater


def _add_to_dependencies(
    auth_repo, branch_name, dependency_name, out_of_band_commit, custom
):

    # add to dependencies.json or update the entry
    dependencies_json = repositoriesdb.load_dependencies_json(auth_repo)

    # if dependencies.json does not exist, initialize it
    if not dependencies_json:
        dependencies_json = {"dependencies": {}}

    dependencies = dependencies_json["dependencies"]
    if dependency_name in dependencies:
        print(f"{dependency_name} already added to dependencies.json. Overwriting")
    dependencies[dependency_name] = {
        "out-of-band-authentication": out_of_band_commit,
        "branch": branch_name,
    }
    if custom:
        dependencies[dependency_name]["custom"] = custom

    # update content of repositories.json before updating targets metadata
    dependencies_path = Path(auth_repo.path, repositoriesdb.DEPENDENCIES_JSON_PATH)
    dependencies_path.parent.mkdir(exist_ok=True)
    Path(dependencies_path).write_text(json.dumps(dependencies_json, indent=4))


@log_on_start(
    DEBUG, "Adding or updating dependency {dependency_name:s}", logger=taf_logger
)
@log_on_end(DEBUG, "Finished adding or updating dependency", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while adding a new dependency: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def add_dependency(
    path: str,
    dependency_name: str,
    branch_name: str,
    out_of_band_commit: str,
    keystore: str,
    dependency_path: Optional[str] = None,
    dependency_url: Optional[str] = None,
    library_dir: Optional[str] = None,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    custom: Optional[Dict] = None,
    prompt_for_keys: Optional[bool] = False,
    commit: Optional[bool] = True,
    push: Optional[bool] = True,
    no_prompt: Optional[bool] = False,
) -> None:
    """
    Add a dependency (an authentication repository) to dependencies.json or update it if it was already added to this file.
    Information that is added to dependencies.json includes out-of-band authentication commit name of the branch which contains
    that commit. It is possible for multiple branches to contain the same commit, so it's important to list branch name as well.
    If repository exists on disk, this function also validates that out_of_brand_commit and branch_name exist and that the
    commit belongs to the specified branch.

    Arguments:
        path: Path to the authentication repository.
        dependency_name: Name of the dependency.
        branch_name: Name of the branch which contains the out-of-band authentication commit.
        out_of_band_commit: SHA of out-of-band authentication commit.
        keystore: Location of the keystore files.
        dependency_path (optional): Path to the dependency repository which is to be added. Can be omitted if dependency_name
        library_dir (optional): Path to the library's root directory. Determined based on the authentication repository's path if not provided.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        custom (optional): Additional data that will be added to dependencies.json if specified.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        push (optional): Flag specifying whether to push to remote.
        no_prompt (optional): Flag specifying whether to ask the user to confirm branch and commit values is they are not specified.
    Side Effects:
        Updates dependencies.json, targets, snapshot and timestamp metadata, writes changes to disk
        and commits them.

    Raises:
        TAFError if dependency exists on disk, but does not contain specified commit and/or branch

    Returns:
        None
    """
    if path is None:
        raise TAFError("Authentication repository's path not provided")

    auth_repo = AuthenticationRepository(path=path)
    if not auth_repo.is_git_repository_root:
        taf_logger.error(f"{path} is not a git repository!")
        return

    dependencies_json = repositoriesdb.load_dependencies_json(auth_repo)
    if dependencies_json is not None and dependency_name in dependencies_json:
        taf_logger.log("NOTICE", f"{dependency_name} already added")
        return
    if library_dir is None:
        library_dir = str(auth_repo.path.parent.parent)

    if dependency_path is not None:
        dependency = GitRepository(path=dependency_path)
    else:
        dependency = GitRepository(library_dir, dependency_name)

    if not dependency.is_git_repository and dependency_url is not None:
        taf_logger.log(
            "NOTICE", f"{dependency.path} does not exist. Cloning from {dependency_url}"
        )
        config = updater.UpdateConfig(
            operation=OperationType.CLONE,
            url=dependency_url,
            path=Path(library_dir, dependency_name),
            library_dir=library_dir,
            strict=False,
            bare=False,
            no_deps=False,
        )  # type: ignore
        try:
            clone_repository(config)
            dependency.default_branch = dependency._determine_default_branch()
        except Exception as e:
            taf_logger.error(f"Dependency clone failed due to error {e}.")
            return

    if dependency.is_git_repository:
        branch_name, out_of_band_commit = _determine_out_of_band_data(
            dependency, branch_name, out_of_band_commit, no_prompt
        )
    else:
        if not no_prompt and not click.confirm(
            "Dependency not on disk. Proceed without validating branch and commit?"
        ):
            return

    if branch_name is None or out_of_band_commit is None:
        raise TAFError(
            "Branch name and out-of-band commit must be specified if repository does not exist on disk"
        )

    _add_to_dependencies(
        auth_repo, branch_name, dependency_name, out_of_band_commit, custom
    )
    commit_msg = git_commit_message("add-dependency", dependency_name=dependency_name)
    register_target_files(
        path=path,
        keystore=keystore,
        commit=commit,
        scheme=scheme,
        auth_repo=auth_repo,
        update_snapshot_and_timestamp=True,
        prompt_for_keys=prompt_for_keys,
        push=push,
        no_commit_warning=True,
        reset_updated_targets_on_error=True,
        commit_msg=commit_msg,
    )


@log_on_start(DEBUG, "Remove dependency {dependency_name:s}", logger=taf_logger)
@log_on_end(DEBUG, "Finished removing dependency", logger=taf_logger)
@log_on_error(
    ERROR,
    "An error occurred while removing a dependency: {e}",
    logger=taf_logger,
    on_exceptions=TAFError,
    reraise=True,
)
@check_if_clean
def remove_dependency(
    path: str,
    dependency_name: str,
    keystore: str,
    scheme: Optional[str] = DEFAULT_RSA_SIGNATURE_SCHEME,
    prompt_for_keys: Optional[bool] = False,
    commit: Optional[bool] = True,
    push: Optional[bool] = True,
) -> None:
    """
    Remove a dependency (an authentication repository) from dependencies.json

    Arguments:
        path: Path to the authentication repository.
        dependency_name: Name of the dependency which should be removed.
        keystore: Location of the keystore files.
        scheme (optional): Signing scheme. Set to rsa-pkcs1v15-sha256 by default.
        prompt_for_keys (optional): Whether to ask the user to enter their key if it is not located inside the keystore directory.
        commit (optional): Indicates if the changes should be committed and pushed automatically.
        push (optional): Flag specifying whether to push to remote
    Side Effects:
        Updates dependencies.json, targets, snapshot and timestamp metadata, writes changes to disk
        and commits them.

    Returns:
        None
    """
    if path is None:
        raise TAFError("Authentication repository's path not provided")

    auth_repo = AuthenticationRepository(path=path)
    if not auth_repo.is_git_repository_root:
        print(f"{path} is not a git repository!")
        return

    # add to dependencies.json or update the entry
    dependencies_json = repositoriesdb.load_dependencies_json(auth_repo)

    if not dependencies_json:
        print("dependencies.json does not exist")
        return

    dependencies = dependencies_json["dependencies"]

    if dependency_name not in dependencies:
        print("Dependency not in dependencies.json")
        return

    dependencies.pop(dependency_name)

    # update content of repositories.json before updating targets metadata
    Path(auth_repo.path, repositoriesdb.DEPENDENCIES_JSON_PATH).write_text(
        json.dumps(dependencies_json, indent=4)
    )

    commit_msg = git_commit_message(
        "remove-dependency", dependency_name=dependency_name
    )

    register_target_files(
        path=path,
        keystore=keystore,
        commit=commit,
        scheme=scheme,
        auth_repo=auth_repo,
        update_snapshot_and_timestamp=True,
        prompt_for_keys=prompt_for_keys,
        push=push,
        no_commit_warning=True,
        reset_updated_targets_on_error=True,
        commit_msg=commit_msg,
    )


def _determine_out_of_band_data(
    dependency: GitRepository,
    branch_name: str,
    out_of_band_commit: str,
    no_prompt: Optional[bool] = False,
):
    """
    Determines values of out-of-band branch and commit as a part of adding a new
    dependency to dependencies.json. If not defined, branch is set to the default branch
    of the repository and commit to the first commit of that branch.
    """
    is_branch_specified = branch_name is not None
    is_commit_specified = out_of_band_commit is not None

    if branch_name is None:
        branch_name = dependency.default_branch
    else:
        if not dependency.branch_exists(branch_name):
            raise TAFError(f"Branch {branch_name} does not exists in {dependency.name}")

    if out_of_band_commit is None:
        out_of_band_commit = dependency.get_first_commit_on_branch(branch_name)
    else:
        try:
            branches = dependency.branches_containing_commit(
                out_of_band_commit, strip_remote=True
            )
        except TAFError:
            raise TAFError("Specified out-of-band authentication commit does not exist")
        if branch_name not in branches:
            raise TAFError(f"Commit {out_of_band_commit} not on branch {branch_name}")

    if not no_prompt and (not is_branch_specified or not is_commit_specified):
        message = f"""
Setting:

    Branch:                             {branch_name}
    out-of-band authentication commit:  {out_of_band_commit}.

Proceed?"""  # noqa: E241
        if not click.confirm(message):
            return None, None

    return branch_name, out_of_band_commit



"""TUF metadata key functions.

"""


from typing import Optional, Tuple, Union

from pathlib import Path
from securesystemslib.signer import (
    SSlibKey,
    CryptoSigner,
    Signer,
    SecretsHandler,
    Signature,
)
from securesystemslib.formats import encode_canonical
from securesystemslib.hash import digest
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME


def generate_rsa_keypair(key_size=3072, password=None) -> Tuple[bytes, bytes]:
    """
    Generate a private-public key pair. Returns the generated keys as bytes in PEM format..
    """
    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=key_size, backend=default_backend()
    )

    # Encrypt the private key if a password is provided
    if password:
        encryption_algorithm = serialization.BestAvailableEncryption(password.encode())
    else:
        encryption_algorithm = serialization.NoEncryption()

    # Serialize private key
    private_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=encryption_algorithm,
    )

    # Get the public key from the private key
    public_key = private_key.public_key()
    # Serialize public key
    public_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    return private_pem, public_pem


def generate_and_write_rsa_keypair(path, key_size, password) -> bytes:
    """
    Generate a private-public key pair and write and save it to files.
    Returns the private key in PEM format.
    """
    if not password:
        password = None
    private_pem, public_pem = generate_rsa_keypair(key_size, password)

    with open(path, "wb") as f:
        f.write(private_pem)

    with open(f"{path}.pub", "wb") as f:
        f.write(public_pem)

    return private_pem


def get_sslib_key_from_value(
    key: str, scheme: str = DEFAULT_RSA_SIGNATURE_SCHEME
) -> SSlibKey:
    """
    Converts a key from its string representation into an SSlibKey object.
    """
    key_val = key.encode()
    crypto_key = load_pem_public_key(key_val, backend=default_backend())
    return _from_crypto(crypto_key, scheme=scheme)


def _get_legacy_keyid(key: SSlibKey) -> str:
    """Computes legacy keyid as hash over an opinionated canonical
    representation of the public key."""
    data = encode_canonical(
        {
            "keytype": key.keytype,
            "scheme": key.scheme,
            "keyval": {"public": key.keyval["public"].strip()},
            "keyid_hash_algorithms": ["sha256", "sha512"],
        }
    ).encode("utf-8")
    hasher = digest("sha256")
    hasher.update(data)
    return hasher.hexdigest()


def _from_crypto(pub: RSAPublicKey, scheme=DEFAULT_RSA_SIGNATURE_SCHEME) -> SSlibKey:
    """Converts pyca/cryptography public key to SSlibKey with default signing
    scheme and legacy keyid."""
    # securesystemslib does not (yet) check if keytype and scheme are compatible
    # https://github.com/secure-systems-lab/securesystemslib/issues/766
    if not isinstance(pub, RSAPublicKey):
        raise ValueError(f"keytype '{type(pub)}' not supported")
    key = SSlibKey.from_crypto(pub, scheme=scheme)
    key.keyid = _get_legacy_keyid(key)
    return key


def load_public_key_from_file(
    path: Union[str, Path], scheme=DEFAULT_RSA_SIGNATURE_SCHEME
) -> SSlibKey:
    """Load SSlibKey from RSA public key file.

    * Expected key file format is SubjectPublicKeyInfo/PEM
    * Signing scheme defaults to 'rsa-pkcs1v15-sha256'
    * Keyid is computed from legacy canonical representation of public key

    """
    # TODO handle scheme
    with open(path, "rb") as f:
        pem = f.read()

    pub = load_pem_public_key(pem)
    return _from_crypto(pub)


def load_signer_from_file(
    path: Path, password: Optional[str] = None, scheme=DEFAULT_RSA_SIGNATURE_SCHEME
) -> CryptoSigner:
    """Load CryptoSigner from RSA private key file.

    * Expected key file format is PKCS8/PEM
    * Signing scheme defaults to 'rsa-pkcs1v15-sha256'
    * Keyid is computed from legacy canonical representation of public key
    * If password is None, the key is expected to be unencrypted

    """
    with open(path, "rb") as f:
        pem = f.read()

    # TODO scheme

    password_encoded = password.encode() if password is not None else None
    priv = load_pem_private_key(pem, password_encoded)
    pub = priv.public_key()
    return CryptoSigner(priv, _from_crypto(pub))


def load_signer_from_pem(pem: bytes, password: Optional[bytes] = None) -> CryptoSigner:
    """Load CryptoSigner from RSA private key file.

    * Expected key file format is PKCS8/PEM
    * Signing scheme defaults to 'rsa-pkcs1v15-sha256'
    * Keyid is computed from legacy canonical representation of public key
    * If password is None, the key is expected to be unencrypted

    """
    priv = load_pem_private_key(pem, password)
    pub = priv.public_key()
    return CryptoSigner(priv, _from_crypto(pub))


class YkSigner(Signer):
    """Signer implementation for Yubikeys.

    Provides a minimal compatibility layer over `taf.yubikey` module functions
    for use with MetadataRepository.

    Attrs:
        public_key: An SSlibkey, whose keyid and signing scheme are the single
                of truth for creating signatures.
        pin_handler: A function, which is called in `sign` and expected to
                return the Yubikey pin.
    """

    _SECRET_PROMPT = "pin"

    def __init__(self, public_key: SSlibKey, pin_handler: SecretsHandler):

        self._public_key = public_key
        self._pin_handler = pin_handler

    @property
    def public_key(self) -> SSlibKey:
        return self._public_key

    @classmethod
    def import_(cls) -> SSlibKey:
        """Import rsa public key from Yubikey.

        * Assigns default signing scheme: "rsa-pkcs1v15-sha256"
        * Raises ValueError, if key on Yubikey is not an rsa key.

        TODO: Consider returning priv key uri along with public key.
        See e.g. `self.from_priv_key_uri` and other `import_` methods on
        securesystemslib signers, e.g. `HSMSigner.import_`.
        """
        # TODO: export pyca/cryptography key to avoid duplicate deserialization
        from taf.yubikey import export_piv_pub_key

        pem = export_piv_pub_key()
        pub = load_pem_public_key(pem)
        return _from_crypto(pub)

    def sign(self, payload: bytes) -> Signature:
        pin = self._pin_handler(self._SECRET_PROMPT)
        # TODO: openlawlibrary/taf#515
        # sig = sign_piv_rsa_pkcs1v15(payload, pin, self.public_key.keyval["public"])
        from taf.yubikey import sign_piv_rsa_pkcs1v15

        sig = sign_piv_rsa_pkcs1v15(payload, pin)
        return Signature(self.public_key.keyid, sig.hex())

    @classmethod
    def from_priv_key_uri(
        cls,
        priv_key_uri: str,
        public_key: SSlibKey,
        secrets_handler: Optional[SecretsHandler] = None,
    ) -> "Signer":
        # TODO: Implement this to better separate public key management
        # (e.g. tuf delegation) and signer configuration from signing. See
        # https://python-securesystemslib.readthedocs.io/en/latest/signer.html
        raise NotImplementedError


def root_signature_provider(signature_dict, key_id, _key, _data):
    """Root signature provider used to return signatures created remotely.

    Args:
        - signature_dict(dict): Dict where key is key_id and value is signature
        - key_id(str): Key id from targets metadata file
        - _key(securesystemslib.formats.RSAKEY_SCHEMA): Key info
        - _data(dict): Data to sign (already signed remotely)

    Returns:
        Dictionary that comforms to `securesystemslib.formats.SIGNATURE_SCHEMA`

    Raises:
        - KeyError: If signature for key_id is not present in signature_dict
    """
    from binascii import hexlify

    return {"keyid": key_id, "sig": hexlify(signature_dict.get(key_id)).decode()}
